from __future__ import annotations

import bisect
import dataclasses
import re
from dataclasses import dataclass, field

from ... import t
from .nodes import TagStack
from .result import Err, Ok, OkT, ResultT, isOk

if t.TYPE_CHECKING:
    from .nodes import ParserNode, StartTag


@dataclass
class ParseConfig:
    algorithm: bool = False
    biblio: bool = False
    cddl: bool = False
    css: bool = False
    dfn: bool = False
    header: bool = False
    idl: bool = False
    macrosInAutolinks: bool = False
    markdownBlock: bool = False
    markdownInline: bool = False
    markdownEscapes: bool = False
    markup: bool = False
    repositoryLinks: bool = False
    macros: dict[str, str] = field(default_factory=dict)
    context: str | None = None
    opaqueElements: set[str] = field(default_factory=lambda: {"pre", "xmp", "script", "style"})

    @staticmethod
    def fromSpec(doc: t.SpecT, context: str | None = None) -> ParseConfig:
        return ParseConfig(
            algorithm="algorithm" in doc.md.markupShorthands,
            biblio="biblio" in doc.md.markupShorthands,
            cddl="cddl" in doc.md.markupShorthands,
            css="css" in doc.md.markupShorthands,
            dfn="dfn" in doc.md.markupShorthands,
            header="http" in doc.md.markupShorthands,
            idl="idl" in doc.md.markupShorthands,
            macrosInAutolinks="macros-in-autolinks" in doc.md.markupShorthands,
            markdownBlock="markdown-block" in doc.md.markupShorthands,
            markdownInline="markdown-inline" in doc.md.markupShorthands,
            markdownEscapes="markdown-escapes" in doc.md.markupShorthands,
            markup="markup" in doc.md.markupShorthands,
            repositoryLinks="repository-links" in doc.md.markupShorthands,
            macros=doc.macros,
            context=context,
            opaqueElements=set(doc.md.opaqueElements),
        )


@dataclass
class Stream:
    _chars: str
    _len: int
    _lineBreaks: list[int]
    startLine: int
    config: ParseConfig
    depth: int = 1
    openEls: TagStack = field(default_factory=TagStack)

    def __init__(self, chars: str, config: ParseConfig, startLine: int = 1, depth: int = 1) -> None:
        if depth > 10:
            msg = "HTML parsing recursed more than 10 levels deep; probably an infinite loop."
            raise RecursionError(msg)
        self._chars = chars
        self._len = len(chars)
        self._lineBreaks = []
        self.startLine = startLine
        self.config = config
        self.depth = depth
        self.openEls = TagStack(opaqueTags=config.opaqueElements)
        for i, char in enumerate(chars):
            if char == "\n":
                self._lineBreaks.append(i)

    def subStream(self, context: str, chars: str, startLine: int = 1) -> Stream:
        newConfig = dataclasses.replace(self.config, context=context)
        return Stream(chars, config=newConfig, startLine=startLine, depth=self.depth + 1)

    def __getitem__(self, key: int) -> str:
        if key < 0 or key >= self._len:
            return ""
        return self._chars[key]

    def slice(self, start: int | None, stop: int | None) -> str:
        if start is not None and start < 0:
            start = 0
        if stop is not None and stop < 0:
            stop = 0
        return self._chars[start:stop]

    def eof(self, index: int) -> bool:
        return index >= self._len

    def __len__(self) -> int:
        return self._len

    @property
    def context(self) -> str | None:
        return self.config.context

    def line(self, index: int) -> int:
        # Zero-based line index
        lineIndex = bisect.bisect_left(self._lineBreaks, index)
        return lineIndex + self.startLine

    def col(self, index: int) -> int:
        lineIndex = bisect.bisect_left(self._lineBreaks, index)
        if lineIndex == 0:
            return index + 1
        startOfCol = self._lineBreaks[lineIndex - 1]
        return index - startOfCol

    def loc(self, index: int) -> str:
        rc = f"{self.line(index)}:{self.col(index)}"
        if self.config.context is None:
            return rc
        return f"{rc} of {self.config.context}"

    def skipTo(self, start: int, text: str) -> ResultT[str]:
        # Skip forward until encountering `text`.
        # Produces the text encountered before this point.
        i = start
        textLen = len(text)
        while not self.eof(i):
            if self.slice(i, i + textLen) == text:
                break
            i += 1
        if self.slice(i, i + textLen) == text:
            return Ok(self.slice(start, i), i)
        else:
            return Err(start)

    def skipToSameLine(self, start: int, text: str) -> ResultT[str]:
        # Skips forward, but no further than the end of the current line.
        # Produces the text encounted before this point.
        i = start
        textLen = len(text)
        while not self.eof(i) and self[i] != "\n":
            if self.slice(i, i + textLen) == text:
                return Ok(self.slice(start, i), i)
            i += 1
        return Err(start)

    def matchRe(self, start: int, pattern: re.Pattern) -> ResultT[re.Match]:
        match = pattern.match(self._chars, start)
        if match:
            return Ok(match, match.end())
        else:
            return Err(start)

    def searchRe(self, start: int, pattern: re.Pattern) -> ResultT[re.Match]:
        match = pattern.search(self._chars, start)
        if match:
            return Ok(match, match.end())
        else:
            return Err(start)

    def skipToNextLine(self, start: int) -> OkT[str]:
        # Skips to the next line.
        # Produces the leftover text on the current line.
        textAfter = self.remainingTextOnLine(start)
        return Ok(textAfter, start + len(textAfter))

    def precedingLinebreakIndex(self, start: int) -> int:
        # Index in self._lineBreaks of the linebreak preceding `start`.
        lineIndex = bisect.bisect_left(self._lineBreaks, start)
        return lineIndex - 1

    def followingLinebreakIndex(self, start: int) -> int:
        # Same but the linebreak after.
        # Note that a newline char is at the end of its line;
        # the actual break in the line is *after* it.
        return bisect.bisect_left(self._lineBreaks, start)

    def currentLineStart(self, start: int) -> int:
        # The index of the first character on the current line.
        lineIndex = self.precedingLinebreakIndex(start)
        if lineIndex == -1:
            return 0
        else:
            return self._lineBreaks[lineIndex] + 1

    def nextLineStart(self, start: int) -> int:
        # The index of the first character on the next line.
        # Returns an OOB index if on the last line.
        lineIndex = self.followingLinebreakIndex(start)
        if lineIndex >= len(self._lineBreaks):
            return len(self._chars)
        else:
            return self._lineBreaks[lineIndex] + 1

    def precedingTextOnLine(self, start: int) -> str:
        # The text on the current line before the start point.
        return self.slice(self.currentLineStart(start), start)

    def remainingTextOnLine(self, start: int) -> str:
        # The text on the current line from the start point on.
        # Includes the newline, if present.
        return self.slice(start, self.nextLineStart(start))

    def observeResult(self, res: ResultT[ParserNode | list[ParserNode]]) -> ResultT[ParserNode | list[ParserNode]]:
        if isOk(res):
            val, _, _ = res
            if isinstance(val, list):
                for node in val:
                    self.observeNode(node)
            else:
                self.observeNode(val)
        return res

    def observeNode(self, node: ParserNode) -> ParserNode:
        self.openEls.update(node)
        return node

    def observeNodes(self, nodes: list[ParserNode]) -> list[ParserNode]:
        for node in nodes:
            self.openEls.update(node)
        return nodes

    def observeShorthandOpen(self, startTag: StartTag, sigils: tuple[str, str]) -> None:
        self.openEls.updateShorthandOpen(startTag, sigils)

    def observeShorthandClose(self, loc: str, startTag: StartTag, sigils: tuple[str, str]) -> None:
        self.openEls.updateShorthandClose(loc, startTag, sigils)

    def cancelShorthandOpen(self, startTag: StartTag, sigils: tuple[str, str]) -> None:
        self.openEls.cancelShorthandOpen(startTag, sigils)

    def inOpaqueElement(self) -> bool:
        return self.openEls.inOpaqueElement()

    def inTagContext(self, tagName: str) -> bool:
        return self.openEls.inTagContext(tagName)



from __future__ import annotations

from collections import Counter

from .. import h, t
from .. import messages as m
from ..translate import _t


def processWptElements(doc: t.SpecT) -> None:
    pathPrefix = doc.md.wptPathPrefix

    atLeastOneElement = False
    atLeastOneVisibleTest = False
    testData = None
    # <wpt> elements
    wptElements = h.findAll("wpt", doc)
    seenTestNames = set()
    prevEl = None
    for el in wptElements:
        if atLeastOneElement is False and el.get("hidden") is None:
            atLeastOneElement = True
        if testData is None:
            testData = loadTestData(doc)
        assert testData is not None
        testNames = testNamesFromEl(el, pathPrefix=pathPrefix)
        for testName in testNames:
            if el.get("skip-existence-check") is None:
                if testName not in testData:
                    m.warn(f"Couldn't find WPT test '{testName}' - did you misspell something?", el=el)
                    continue
            seenTestNames.add(testName)
            if atLeastOneVisibleTest is False and el.get("hidden") is None:
                atLeastOneVisibleTest = True
        testNames = [x for x in testNames if x in seenTestNames]
        if el.get("hidden") is not None:
            h.removeNode(el)
        else:
            title = el.get("title")
            titleLang = el.get("lang")
            titleDir = el.get("dir")
            if prevEl is not None and prevEl.getnext() is el and (prevEl.tail is None or prevEl.tail.strip() == ""):
                appendTestList(prevEl, testNames, testData, title, titleLang, titleDir)
                h.removeNode(el)
            else:
                createHTML(doc, el, testNames, testData, title, titleLang, titleDir)
                prevEl = el

    # <wpt-rest> elements
    wptRestElements = h.findAll("wpt-rest", doc)
    if wptRestElements:
        if testData is None:
            testData = loadTestData(doc)
        assert testData is not None
        if len(wptRestElements) > 1:
            m.die(f"Only one <wpt-rest> element allowed per document, you have {len(wptRestElements)}.")
            wptRestElements = wptRestElements[0:1]
        else:
            localPrefix = wptRestElements[0].get("pathprefix")
            if localPrefix is not None:
                pathPrefix = localPrefix
            if pathPrefix is None:
                m.die("Can't use <wpt-rest> without either a pathprefix='' attribute or a 'WPT Path Prefix' metadata.")
                return
            prefixedNames = [p for p in testData if prefixInPath(pathPrefix, p) and p not in seenTestNames]
            if len(prefixedNames) == 0:
                m.die(f"Couldn't find any tests with the path prefix '{pathPrefix}'.")
                return
            atLeastOneElement = True
            atLeastOneVisibleTest = True
            createHTML(doc, wptRestElements[0], prefixedNames, testData)
            m.warn(
                "<wpt-rest> is intended for debugging only. Move the tests to <wpt> elements next to what they're testing.",
            )
    else:
        if pathPrefix:
            if testData is None:
                testData = loadTestData(doc)
            checkForOmittedTests(pathPrefix, testData, seenTestNames)

    if atLeastOneVisibleTest:
        if pathPrefix and pathPrefix.startswith("https:"):
            dataPaths = [pathPrefix]
        else:
            dataPaths = commonPathPrefixes(seenTestNames)

        testSuiteUrl = None
        if pathPrefix and not pathPrefix.startswith("https:"):
            pathPrefix = "/" + pathPrefix.strip("/") + "/"
            testSuiteUrl = f"https://wpt.fyi/results{pathPrefix}"
        elif not pathPrefix:
            testSuiteUrl = guessTestSuiteUrl(seenTestNames)
        if testSuiteUrl:
            doc.md.otherMetadata.setdefault(_t("Test Suite"), []).append(
                h.E.dd(
                    {"class": "wpt-overview"},
                    h.E.a(
                        {"href": testSuiteUrl},
                        testSuiteUrl,
                    ),
                ),
            )

    if doc.md.wptDisplay != "none":
        if atLeastOneElement:
            # Empty <wpt> blocks still need styles
            doc.extraJC.addWptCSS()
        if atLeastOneVisibleTest:
            # But I only need script if there's actually some tests.
            doc.extraJC.addWpt(dataPaths)


def createHTML(
    doc: t.SpecT,
    blockEl: t.ElementT,
    testNames: list[str],
    testData: dict[str, str],
    title: str | None = None,
    titleLang: str | None = None,
    titleDir: str | None = None,
) -> None:
    if doc.md.wptDisplay == "none":
        h.removeNode(blockEl)
    elif doc.md.wptDisplay in ("inline", "open", "closed"):
        blockEl.tag = "details"
        h.addClass(doc, blockEl, "wpt-tests-block")
        h.removeAttr(blockEl, "pathprefix")
        h.removeAttr(blockEl, "title")
        blockEl.set("lang", "en")
        blockEl.set("dir", "ltr")
        if doc.md.wptDisplay in ("open", "inline"):
            blockEl.set("open", "")
        h.clearContents(blockEl)
        testSummaryEl = h.E.summary(_t("Tests"))
        h.appendChild(blockEl, testSummaryEl)
        appendTestList(blockEl, testNames, testData, title, titleLang, titleDir)
    else:
        m.die("Programming error, uncaught WPT Display value in createHTML.")


def appendTestList(
    blockEl: t.ElementT,
    testNames: list[str],
    testData: dict[str, str],
    title: str | None = None,
    titleLang: str | None = None,
    titleDir: str | None = None,
) -> None:
    if title:
        titleEl = h.E.p(
            {
                "lang": titleLang,
                "dir": titleDir,
            },
            h.parseHTML(title),
        )
        h.appendChild(blockEl, titleEl)
    testListEl = h.E.ul({"class": "wpt-tests-list"})
    h.appendChild(blockEl, testListEl)
    for testName in testNames:
        if testName not in testData:
            if testName.startswith("https:"):
                # Assume it's a custom URL, just link to it.
                singleTestEl = h.E.li(
                    {"class": "wpt-test"},
                    h.E.a(
                        {
                            "href": testName,
                            "class": "wpt-name",
                            # Since the name is a URL and might be long,
                            # let it break freely.
                            "style": "word-break: break-all",
                        },
                        testName,
                    ),
                )
                h.appendChild(testListEl, singleTestEl)
                continue
            else:
                m.warn(f"Cannot find '{testName}' in the test data.")
                continue
        if ".https." in testName or ".serviceworker." in testName:
            liveTestScheme = "https"
        else:
            liveTestScheme = "http"
        _, _, lastNameFragment = testName.rpartition("/")
        testType = testData[testName]
        if testType in ["crashtest", "print-reftest", "reftest", "testharness"]:
            singleTestEl = h.E.li(
                {"class": "wpt-test"},
                h.E.a(
                    {
                        "title": testName,
                        "href": "https://wpt.fyi/results/" + testName,
                        "class": "wpt-name",
                    },
                    lastNameFragment,
                ),
                " ",
                h.E.a(
                    {
                        "href": f"{liveTestScheme}://wpt.live/{testName}",
                        "class": "wpt-live",
                    },
                    h.E.small(_t("(live test)")),
                ),
                " ",
                h.E.a(
                    {
                        "href": "https://github.com/web-platform-tests/wpt/blob/master/" + testName,
                        "class": "wpt-source",
                    },
                    h.E.small(_t("(source)")),
                ),
            )
        elif testType in ["manual", "visual"]:
            singleTestEl = h.E.li(
                {"class": "wpt-test"},
                h.E.span({"class": "wpt-name"}, lastNameFragment, f" ({testType} test) "),
                h.E.a(
                    {
                        "href": "https://github.com/web-platform-tests/wpt/blob/master/" + testName,
                        "class": "wpt-source",
                    },
                    h.E.small(_t("(source)")),
                ),
            )
        elif testType in ["wdspec"]:
            singleTestEl = h.E.li(
                {"class": "wpt-test"},
                h.E.a(
                    {
                        "href": "https://wpt.fyi/results/" + testName,
                        "class": "wpt-name",
                    },
                    lastNameFragment,
                ),
                " ",
                h.E.a(
                    {
                        "href": "https://github.com/web-platform-tests/wpt/blob/master/" + testName,
                        "class": "wpt-source",
                    },
                    h.E.small(_t("(source)")),
                ),
            )
        else:
            m.warn(
                f"Programming error, the test {testName} is of type {testType}, which I don't know how to render. Please report this!",
            )
            continue
        h.appendChild(testListEl, singleTestEl)
    if title:
        h.appendChild(blockEl, h.E.hr())


def testNamesFromEl(el: t.ElementT, pathPrefix: str | None = None) -> list[str]:
    testNames = []
    localPrefix = el.get("pathprefix")
    if localPrefix is not None:
        pathPrefix = localPrefix
    for name in [x.strip() for x in h.textContent(el).split("\n")]:
        if name == "":
            continue
        testName = prefixPlusPath(pathPrefix, name)
        testNames.append(testName)
    return testNames


def prefixPlusPath(prefix: str | None, path: str) -> str:
    # Join prefix to path, normalizing slashes
    if path.startswith("/"):
        return path[1:]
    prefix = normalizePathSegment(prefix)
    if prefix is None:
        return path
    return prefix + path


def prefixInPath(prefix: str | None, path: str) -> bool:
    if prefix is None:
        return False
    return path.startswith(normalizePathSegment(prefix))


@t.overload
def normalizePathSegment(pathSeg: str) -> str: ...


@t.overload
def normalizePathSegment(pathSeg: None) -> None: ...


def normalizePathSegment(pathSeg: str | None) -> str | None:
    # No slash at front, yes slash at end
    if pathSeg is None:
        return None
    if pathSeg.startswith("/"):
        pathSeg = pathSeg[1:]
    if not pathSeg.endswith("/"):
        pathSeg += "/"
    return pathSeg


def checkForOmittedTests(pathPrefix: str, testData: dict[str, str], seenTestNames: set[str]) -> None:
    unseenTests = []
    for testPath in testData:
        if ".tentative." in testPath:
            continue
        if prefixInPath(pathPrefix, testPath):
            if testPath not in seenTestNames:
                unseenTests.append(testPath)
    if unseenTests:
        numTests = len(unseenTests)
        m.warn(
            f"There are {numTests} WPT tests underneath your path prefix '{pathPrefix}' that aren't in your document and must be added. (Use a <wpt hidden> if you don't actually want them in your document.)\n"
            + "\n".join("  " + path for path in sorted(unseenTests)),
        )


def loadTestData(doc: t.SpecT) -> dict[str, str]:
    paths = {}
    for line in doc.dataFile.fetch("wpt-tests.txt", str=True).split("\n")[1:]:
        testType, _, testPath = line.strip().partition(" ")
        paths[testPath] = testType
    return paths


def xor(a: t.Any, b: t.Any) -> bool:
    return bool(a) != bool(b)


def commonPathPrefixes(paths: t.Iterable[str]) -> list[str]:
    # Split the paths into segments, and drop the filename
    splitPaths = [x.strip("/").split("/")[:-1] for x in paths]
    # Only record the first two segments, that's good enough
    # for limiting the data slurped.
    oneSegs = set()
    twoSegs = set()
    for pathSegs in splitPaths:
        prefix = pathSegs[0:2]
        if len(prefix) == 1:
            oneSegs.add("/" + prefix[0] + "/")
        else:
            twoSegs.add("/" + "/".join(prefix) + "/")
    # Filter out any 2-length segments already covered by a 1-length seg
    for short in oneSegs:
        twoSegs = {x for x in twoSegs if not x.startswith(short)}
    return list(oneSegs) + list(twoSegs)


def guessTestSuiteUrl(paths: t.Iterable[str]) -> str:
    # Do the same transforms as commonPathPrefixes, but then
    # just count the number of paths under each,
    # and return the one that wins.
    splitPaths = [x.strip("/").split("/")[:-1] for x in paths]
    prefixCounter: Counter[str] = Counter()
    for pathSegs in splitPaths:
        prefix = "/" + "/".join(pathSegs[0:2]) + "/"
        prefixCounter[prefix] += 1
    bestGuess = prefixCounter.most_common(1)[0][0]
    return f"https://wpt.fyi/results{bestGuess}"



from __future__ import annotations

import contextlib
import dataclasses
import io
import json
import os
import sys
from collections import Counter

import lxml.html

from . import t

MESSAGE_LEVELS = {
    "everything": 0,
    "message": 1,
    "lint": 2,
    "warning": 3,
    "link-error": 4,
    "fatal": 5,
    "nothing": 6,
}

DEATH_TIMING = [
    "early",  # die as soon as the first disallowed error occurs
    "late",  # die at the end of processing
]

PRINT_MODES = [
    "plain",
    "console",
    "markup",
    "json",
]


@dataclasses.dataclass
class MessageOptions:
    msg: str
    lineNum: int | str | None = None
    el: t.ElementT | None = None

    def die(self) -> None:
        die(**dataclasses.asdict(self))

    def linkerror(self) -> None:
        linkerror(**dataclasses.asdict(self))

    def lint(self) -> None:
        lint(**dataclasses.asdict(self))

    def warn(self) -> None:
        warn(**dataclasses.asdict(self))


@dataclasses.dataclass()
class MessagesState:
    # What message category (or higher) to stop processing on
    dieOn: str = "fatal"
    # When to stop processing when an error that trips failure happens
    dieWhen: str = "late"
    # What message category (or higher) to print
    printOn: str = "everything"
    # Suppress *all* categories, *plus* the final success/fail message
    silent: bool = False
    printMode: str = "console"
    asciiOnly: bool = False
    fh: t.TextIO = sys.stdout
    seenMessages: set[str | tuple[str, str]] = dataclasses.field(default_factory=set)
    categoryCounts: Counter[str] = dataclasses.field(default_factory=Counter)
    closed: bool = False

    def __post_init__(self) -> None:
        if isinstance(self.fh, io.TextIOWrapper):
            self.fh.reconfigure(encoding="utf-8")

    def record(self, category: str, message: str | tuple[str, str]) -> None:
        self.categoryCounts[category] += 1
        self.seenMessages.add(message)

    def replace(self, **kwargs: t.Any) -> MessagesState:
        return dataclasses.replace(self, seenMessages=set(), categoryCounts=Counter(), **kwargs)

    def shouldDie(self, category: str, timing: str = "early") -> bool:
        if self.dieWhen == "never":
            return False
        if self.dieWhen == "late" and timing == "early":
            return False
        deathLevel = MESSAGE_LEVELS[self.dieOn]
        queriedLevel = MESSAGE_LEVELS[category]
        return queriedLevel >= deathLevel

    def shouldPrint(self, category: str) -> bool:
        if self.silent:
            return False
        if category in ("success", "failure"):
            return True
        printLevel = MESSAGE_LEVELS[self.printOn]
        queriedLevel = MESSAGE_LEVELS[category]
        return queriedLevel >= printLevel

    @staticmethod
    def categoryName(categoryNum: int) -> str:
        assert categoryNum >= 0
        if categoryNum >= len(MESSAGE_LEVELS):
            return "nothing"
        return list(MESSAGE_LEVELS.keys())[categoryNum]


state = MessagesState()


def wrappedOutput() -> bool:
    return state.printMode in ("json", "markup")


def p(msg: str | tuple[str, str]) -> None:
    if isinstance(msg, tuple):
        msg, ascii = msg
    else:
        ascii = msg.encode("ascii", "replace").decode()
    if state.asciiOnly:
        msg = ascii
    try:
        print(msg, file=state.fh)
    except UnicodeEncodeError:
        if ascii is not None:
            print(ascii, file=state.fh)
        else:
            warning = formatMessage(
                "warning",
                "Your console does not understand Unicode.\n  Messages may be slightly corrupted.",
            )
            if warning not in state.seenMessages:
                print(warning, file=state.fh)
                state.record("warning", warning)
            print(msg.encode("ascii", "xmlcharrefreplace"), file=state.fh)


def getLineNum(lineNum: str | int | t.ElementT | None = None, el: t.ElementT | None = None) -> str | int | None:
    if isinstance(lineNum, str):
        return lineNum
    elif isinstance(lineNum, int):
        return str(lineNum)
    elif lineNum is not None:
        return getLineNum(el=lineNum)
    elif el is not None and el.get("bs-line-number"):
        s = el.get("bs-line-number", "")
        context = el.get("bs-parse-context", None)
        if context:
            s += " of " + context
        return s
    return None


def printOpener() -> None:
    if state.printMode == "json":
        p("[")
    elif state.printMode == "markup":
        p("<bikeshed-output>")
    else:
        pass


def printCloser() -> None:
    if state.closed:
        return
    state.closed = True

    if state.categoryCounts["success"] == 0 and state.categoryCounts["failure"] == 0:
        # If something closes without actually going through success()/failure() first,
        # insert a dummy success message.
        # (Partially just to record *something*, but also to make sure that, say, JSON
        #  formats properly wrt commas.)
        if state.printMode in ("json", "markup") and state.shouldPrint("success"):
            p(formatMessage("success", "Success"))

    if state.printMode == "json":
        p("]")
    elif state.printMode == "markup":
        p("</bikeshed-output>")
    else:
        pass


def die(msg: str, el: t.ElementT | None = None, lineNum: str | int | None = None) -> None:
    lineNum = getLineNum(lineNum, el)
    formattedMsg = formatMessage("fatal", msg, lineNum=lineNum)
    if formattedMsg not in state.seenMessages:
        state.record("fatal", formattedMsg)
        if state.shouldPrint("fatal"):
            p(formattedMsg)
    if state.shouldDie("fatal"):
        errorAndExit()


def linkerror(msg: str, el: t.ElementT | None = None, lineNum: str | int | None = None) -> None:
    lineNum = getLineNum(lineNum, el)
    suffix = ""
    if el is not None:
        if el.get("bs-autolink-syntax"):
            suffix = "\n" + t.cast(str, el.get("bs-autolink-syntax"))
        else:
            suffix = "\n" + lxml.html.tostring(el, with_tail=False, encoding="unicode")
    formattedMsg = formatMessage("link", msg + suffix, lineNum=lineNum)
    if formattedMsg not in state.seenMessages:
        state.record("link-error", formattedMsg)
        if state.shouldPrint("link-error"):
            p(formattedMsg)
    if state.shouldDie("link-error"):
        errorAndExit()


def lint(msg: str, el: t.ElementT | None = None, lineNum: str | int | t.ElementT | None = None) -> None:
    lineNum = getLineNum(lineNum, el)
    suffix = ""
    if el is not None:
        if el.get("bs-autolink-syntax"):
            suffix = "\n" + t.cast(str, el.get("bs-autolink-syntax"))
        else:
            suffix = "\n" + lxml.html.tostring(el, with_tail=False, encoding="unicode")
    formattedMsg = formatMessage("lint", msg + suffix, lineNum=lineNum)
    if formattedMsg not in state.seenMessages:
        state.record("lint", formattedMsg)
        if state.shouldPrint("lint"):
            p(formattedMsg)
    if state.shouldDie("lint"):
        errorAndExit()


def warn(msg: str, el: t.ElementT | None = None, lineNum: str | int | None = None) -> None:
    if lineNum is None and el is not None and el.get("bs-line-number"):
        lineNum = el.get("bs-line-number")
    formattedMsg = formatMessage("warning", msg, lineNum=lineNum)
    if formattedMsg not in state.seenMessages:
        state.record("warning", formattedMsg)
        if state.shouldPrint("warning"):
            p(formattedMsg)
    if state.shouldDie("warning"):
        errorAndExit()


def say(msg: str) -> None:
    if state.shouldPrint("message"):
        p(formatMessage("message", msg))


def success(msg: str) -> None:
    if state.shouldPrint("success"):
        formattedMsg = formatMessage("success", msg)
        p(formattedMsg)
        state.record("success", formattedMsg)
        printCloser()


def failure(msg: str) -> None:
    if state.shouldPrint("failure"):
        formattedMsg = formatMessage("failure", msg)
        p(formattedMsg)
        state.record("failure", formattedMsg)
        printCloser()


def retroactivelyCheckErrorLevel(level: str | None = None, timing: str = "early") -> bool:
    if level is None:
        level = state.dieOn
    for levelName, msgCount in state.categoryCounts.items():
        if msgCount > 0 and state.shouldDie(levelName, timing):
            errorAndExit()
    return True


def printColor(text: str, color: str = "white", *styles: str) -> str:
    if state.printMode == "console":
        colorsConverter = {
            "black": 30,
            "red": 31,
            "green": 32,
            "yellow": 33,
            "blue": 34,
            "magenta": 35,
            "cyan": 36,
            "light gray": 37,
            "dark gray": 90,
            "light red": 91,
            "light green": 92,
            "light yellow": 93,
            "light blue": 94,
            "light magenta": 95,
            "light cyan": 96,
            "white": 97,
        }
        stylesConverter = {
            "normal": 0,
            "bold": 1,
            "bright": 1,
            "dim": 2,
            "underline": 4,
            "underlined": 4,
            "blink": 5,
            "reverse": 7,
            "invert": 7,
            "hidden": 8,
        }

        colorNum = colorsConverter[color.lower()]
        styleNum = ";".join(str(stylesConverter[style.lower()]) for style in styles)
        return f"\033[{styleNum};{colorNum}m{text}\033[0m"
    return text


def formatMessage(type: str, text: str, lineNum: str | int | None = None) -> str | tuple[str, str]:
    if state.printMode == "markup":
        text = text.replace("<", "&lt;")
        if type == "fatal":
            return f"  <fatal>{text}</fatal>"
        if type == "link":
            return f"  <linkerror>{text}</linkerror>"
        if type == "lint":
            return f"  <lint>{text}</lint>"
        if type == "warning":
            return f"  <warning>{text}</warning>"
        if type == "message":
            return f"  <message>{text}</message>"
        if type == "success":
            return f"  <final-success>{text}</final-success>"
        if type == "failure":
            return f"  <final-failure>{text}</final-failure>"
    elif state.printMode == "json":
        msg = {"lineNum": lineNum, "messageType": type, "text": text}
        jsonText = "  " + json.dumps(msg)
        if type not in ("success", "failure"):
            jsonText += ", "
        return jsonText

    if type == "message":
        return text
    if type == "success":
        return (
            printColor(" ✔ ", "green", "invert") + " " + text,
            printColor("YAY", "green", "invert") + " " + text,
        )
    if type == "failure":
        return (
            printColor(" ✘ ", "red", "invert") + " " + text,
            printColor("ERR", "red", "invert") + " " + text,
        )
    if type == "fatal":
        headingText = "FATAL ERROR"
        color = "red"
    elif type == "link":
        headingText = "LINK ERROR"
        color = "yellow"
    elif type == "lint":
        headingText = "LINT"
        color = "yellow"
    elif type == "warning":
        headingText = "WARNING"
        color = "light cyan"
    if lineNum is not None:
        headingText = f"LINE {lineNum}"
    return printColor(headingText + ":", color, "bold") + " " + text  # pylint: disable=possibly-used-before-assignment


def errorAndExit() -> None:
    failure("Did not generate, due to errors exceeding the allowed error level.")
    sys.exit(2)


@contextlib.contextmanager
def withMessageState(
    fh: str | t.TextIO,
    **kwargs: t.Any,
) -> t.Generator[t.TextIO, None, None]:
    if isinstance(fh, str):
        fhIsTemporary = True
        fh = open(fh, "w", encoding="utf-8")
        assert isinstance(fh, io.TextIOWrapper)
    else:
        fhIsTemporary = False
    global state
    oldState = state
    state = oldState.replace(fh=fh, **kwargs)
    try:
        yield fh
    finally:
        state = oldState
        if fhIsTemporary:
            fh.close()


@contextlib.contextmanager
def messagesSilent() -> t.Generator[t.TextIO, None, None]:
    fh = open(os.devnull, "w", encoding="utf-8")
    global state
    oldState = state
    state = oldState.replace(fh=fh)
    try:
        yield fh
    finally:
        state = oldState
        fh.close()



# pylint: skip-file
from __future__ import annotations

import enum
import io
import os
import re

from ... import constants, t
from ... import messages as m
from .nodes import (
    Comment,
    Doctype,
    EndTag,
    ParserNode,
    RawElement,
    RawText,
    SafeText,
    SelfClosedTag,
    StartTag,
    Text,
)
from .parser import POSSIBLE_NODE_START_CHARS, nodesFromStream
from .stream import ParseConfig, Stream


def nodesFromHtml(data: str, config: ParseConfig, startLine: int = 1) -> t.Generator[ParserNode, None, None]:
    s = Stream(data, startLine=startLine, config=config)
    yield from nodesFromStream(s, 0)


def initialDocumentParse(text: str, config: ParseConfig, startLine: int = 1) -> list[ParserNode]:
    # Just do a document parse.
    # This will add `bs-line-number` attributes,
    # normalize any difficult shorthands
    # (ones that look like tags, or that contain raw text),
    # and blank out comments.

    return list(nodesFromHtml(text, config, startLine=startLine))


def strFromNodes(nodes: t.Iterable[ParserNode], withIlcc: bool = False) -> str:
    strs = []
    ilcc = constants.incrementLineCountChar
    dlcc = constants.decrementLineCountChar
    for node in nodes:
        if isinstance(node, Comment):
            # Serialize comments as a standardized, recognizable sequence
            # so Markdown processing can ignore them better.
            strs.append(constants.bsComment)
            if withIlcc:
                strs.append(ilcc * node.data.count("\n"))
            continue
        s = str(node)
        if withIlcc:
            diff = node.height - s.count("\n")
            if diff > 0:
                s += ilcc * diff
            elif diff < 0:
                s += dlcc * -diff
        strs.append(s)
    return "".join(strs)


def linesFromNodes(nodes: t.Iterable[ParserNode]) -> list[str]:
    return strFromNodes(nodes).split("\n")


def debugNodes(nodes: t.Iterable[ParserNode]) -> list[ParserNode]:
    nodes = list(nodes)
    for node in nodes:
        print("------")  # noqa: T201
        print(repr(node))  # noqa: T201
        print(repr(strFromNodes([node], withIlcc=True)))  # noqa: T201
    return nodes


def parseLines(textLines: list[str], config: ParseConfig, startLine: int = 1) -> list[str]:
    # Runs a list of lines thru the parser,
    # returning another list of lines.

    if len(textLines) == 0:
        return textLines
    endingWithNewline = textLines[0].endswith("\n")
    if endingWithNewline:
        text = "".join(textLines)
    else:
        text = "\n".join(textLines)
    parsedLines = strFromNodes(nodesFromHtml(text, config, startLine=startLine)).split("\n")
    if endingWithNewline:
        parsedLines = [x + "\n" for x in parsedLines]

    return parsedLines


def parseText(text: str, config: ParseConfig, startLine: int = 1) -> str:
    # Just runs the text thru the parser.
    return strFromNodes(nodesFromHtml(text, config, startLine=startLine))


def parseTitle(text: str, config: ParseConfig, startLine: int = 1) -> str:
    # Parses the text, but removes any tags from the content,
    # as they'll just show up as literal text in <title>.
    nodes = nodesFromHtml(text, config, startLine=startLine)
    return strFromNodes(n for n in nodes if isinstance(n, Text))



# pylint: disable=wrong-import-position

from __future__ import annotations

import os
import platform
import subprocess
import sys


def verify_python_version() -> None:
    if sys.version_info < (3,):
        print(
            """Bikeshed has updated to Python 3, but you are trying to run it with
    Python {}. For instructions on upgrading, please check:
    https://speced.github.io/bikeshed/#installing""".format(
                platform.python_version(),
            ),
        )
        sys.exit(1)

    if sys.version_info < (3, 9):
        print(
            """Bikeshed now requires Python 3.9 or higher; you are on {}.
    For instructions on how to set up a pyenv with 3.9, see:
    https://speced.github.io/bikeshed/#installing""".format(
                platform.python_version(),
            ),
        )
        sys.exit(1)


verify_python_version()


def verify_requirements() -> None:
    try:
        subprocess.check_output([sys.executable, "-m", "pip", "check", "bikeshed"])  # noqa: S603
    except subprocess.CalledProcessError as e:
        print("ERROR: Broken or incomplete manual installation.")
        print(str(e.output, encoding="utf-8"))
        requirements_file_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
            "requirements.txt",
        )
        print(f'Run "pip3 install -r {requirements_file_path}" to complete installation')
        print("Meanwhile, attempting to run Bikeshed anyway...")


verify_requirements()

from . import (
    config,
    update,
)
from .cli import main
from .Spec import Spec



import re

from pygments import formatters, style, token

# dark
text = "#657b83"
bg = "#002b36"

# light
# text = "#839496"
# bg = "#fdf6e3"

# both
yellow = "#b58900"
orange = "#cb4b16"
red = "#dc322f"
magenta = "#d33682"
violet = "#6c71c4"
blue = "#268bd2"
cyan = "#2aa198"
green = "#859900"


class PrismStyle(style.Style):
    default_style = text
    styles = {
        token.Name: blue,
        token.Name.Tag: yellow,
        token.Name.Builtin: "noinherit",
        token.Name.Variable: orange,
        token.Name.Other: "noinherit",
        token.Operator: text,
        token.Punctuation: text,
        token.Keyword: magenta,
        token.Literal: text,
        token.Literal.Number: text,
        token.Literal.String: violet,
        token.Comment: cyan,
    }


css = formatters.HtmlFormatter(style=PrismStyle).get_style_defs(".c")  # pylint: disable=no-member
css = re.sub(r"\.c \.(\w+)", r"c-[\1]", css)
print(css)  # noqa: T201



from __future__ import annotations

import json
from collections import OrderedDict
from datetime import datetime

from .. import h, t
from .. import messages as m
from ..translate import _t


def addCanIUsePanels(doc: t.SpecT) -> list[t.ElementT]:
    # Constructs "Can I Use panels" which show a compatibility data summary
    # for a term's feature.
    if not doc.md.includeCanIUsePanels:
        return []

    canIUseData = CanIUseManager(dataFile=doc.dataFile)

    lastUpdated = datetime.utcfromtimestamp(canIUseData.updated).date().isoformat()

    # e.g. 'iOS Safari' -> 'ios_saf'
    classFromBrowser = canIUseData.agents

    panels = []
    elements = h.findAll("[caniuse]", doc)
    if not elements:
        return []
    validateCanIUseURLs(doc, canIUseData, elements)
    for dfn in elements:
        dfnId = dfn.get("id")
        if not dfnId:
            m.die(f"Elements with `caniuse` attribute need to have an ID as well. Got:\n{h.serializeTag(dfn)}", el=dfn)
            continue

        featId = dfn.get("caniuse")
        if not featId:
            continue
        if not canIUseData.hasFeature(featId):
            m.die(f"Unrecognized Can I Use feature ID: {featId}", el=dfn)
            continue
        del dfn.attrib["caniuse"]
        featId = featId.lower()
        feature = canIUseData.getFeature(featId)

        h.addClass(doc, dfn, "caniuse-paneled")
        panel = canIUsePanelFor(
            id=featId,
            data=feature,
            update=lastUpdated,
            classFromBrowser=classFromBrowser,
        )
        panel.set("data-anno-for", dfnId)
        h.appendChild(doc.body, panel)
        panels.append(panel)

    if panels:
        doc.extraJC.addCiu()

    return panels


def canIUsePanelFor(id: str, data: t.JSONT, update: str, classFromBrowser: dict[str, str]) -> t.ElementT:
    panel = h.E.details(
        {"class": "caniuse-status unpositioned", "data-deco": ""},
        h.E.summary({}, "CanIUse"),
    )
    mainPara = h.E.p({"class": "support"}, h.E.b({}, _t("Support:")))
    h.appendChild(panel, mainPara)
    for browser, support in data["support"].items():
        statusCode = support[0]
        if statusCode == "u":
            continue
        minVersion = support[2:]
        h.appendChild(
            mainPara,
            browserCompatSpan(classFromBrowser[browser], browser, statusCode, minVersion),
        )
    h.appendChild(
        panel,
        h.E.p(
            {"class": "caniuse"},
            _t("Source: "),
            h.E.a({"href": "https://caniuse.com/#feat=" + id}, "caniuse.com"),
            _t(" as of {date}").format(date=update),
        ),
    )
    return panel


def browserCompatSpan(
    browserCodeName: str,
    browserFullName: str,
    statusCode: str,
    minVersion: str | None = None,
) -> t.ElementT:
    if statusCode == "n" or minVersion is None:
        minVersion = "None"
    elif minVersion == "all":
        minVersion = "All"
    else:
        minVersion = minVersion + "+"
    # browserCodeName: e.g. and_chr, ios_saf, ie, etc...
    # browserFullName: e.g. "Chrome for Android"
    statusClass = {"y": "yes", "n": "no", "a": "partial"}[statusCode]
    outer = h.E.span({"class": browserCodeName + " " + statusClass})
    if statusCode == "a":
        h.appendChild(outer, h.E.span({}, h.E.span({}, browserFullName, _t(" (limited)"))))
    else:
        h.appendChild(outer, h.E.span({}, browserFullName))
    h.appendChild(outer, h.E.span({}, minVersion))
    return outer


def validateCanIUseURLs(doc: t.SpecT, canIUseData: CanIUseManager, elements: list[t.ElementT]) -> None:
    # First, ensure that each Can I Use URL shows up at least once in the data;
    # otherwise, it's an error to be corrected somewhere.
    urlFeatures = set()
    for url in doc.md.canIUseURLs:
        sawTheURL = False
        for featureID, featureUrl in canIUseData.urlFromFeature.items():
            if featureUrl.startswith(url):
                sawTheURL = True
                urlFeatures.add(featureID)
        if not sawTheURL and url not in doc.md.ignoreCanIUseUrlFailure:
            m.warn(
                f"The Can I Use URL '{url}' isn't associated with any of the Can I Use features."
                + "Please check Can I Use for the correct spec url, and either correct your spec or correct Can I Use."
                + "If the URL is correct and you'd like to keep it in pre-emptively, add the URL to a 'Ignore Can I Use URL Failure' metadata.",
            )

    # Second, ensure that every feature in the data corresponding to one of the listed URLs
    # has a corresponding Can I Use entry in the document;
    # otherwise, you're missing some features.
    docFeatures = set()
    for el in elements:
        canIUseAttr = el.get("caniuse")
        assert canIUseAttr is not None
        featureID = canIUseAttr.lower()
        docFeatures.add(featureID)

    unusedFeatures = urlFeatures - docFeatures
    if unusedFeatures:
        featureList = "\n".join(" * {0} - https://caniuse.com/#feat={0}".format(x) for x in sorted(unusedFeatures))
        m.warn(
            f"The following Can I Use features are associated with your URLs, but don't show up in your spec:\n{featureList}",
        )


class CanIUseManager:
    def __init__(self, dataFile: t.DataFileRequester) -> None:
        self.dataFile = dataFile
        data = json.loads(
            self.dataFile.fetch("caniuse", "data.json", str=True),
            object_pairs_hook=OrderedDict,
        )
        self.updated = data["updated"]
        self.agents = data["agents"]
        self.urlFromFeature = data["features"]
        self.features: t.JSONT = {}

    def hasFeature(self, featureName: str) -> bool:
        return featureName in self.urlFromFeature

    def getFeature(self, featureName: str) -> t.JSONT:
        if featureName in self.features:
            return t.cast("t.JSONT", self.features[featureName])
        if not self.hasFeature(featureName):
            return {}
        data = json.loads(
            self.dataFile.fetch("caniuse", f"feature-{featureName}.json", str=True),
            object_pairs_hook=OrderedDict,
        )
        self.features[featureName] = data
        return t.cast("t.JSONT", data)



# pylint: skip-file
# Module for holding types, for easy importing into the rest of the codebase
from __future__ import annotations

import sys

# The only things that should be available during runtime.
from typing import TYPE_CHECKING, Generic, TypedDict, TypeVar, cast, overload

# Only available in 3.11, so stub them out for earlier versions
if sys.version_info >= (3, 11):
    from typing import assert_never, assert_type
else:
    from typing_extensions import assert_never, assert_type


if TYPE_CHECKING:
    from typing import (
        AbstractSet,
        Any,
        AnyStr,
        Awaitable,
        Callable,
        Collection,
        DefaultDict,
        Deque,
        FrozenSet,
        Generator,
        Iterable,
        Iterator,
        Literal,
        Mapping,
        MutableMapping,
        MutableSequence,
        NamedTuple,
        NewType,
        Protocol,
        Sequence,
        TextIO,
        Type,
        TypeAlias,
        TypeGuard,
    )

    from _typeshed import SupportsKeysAndGetItem
    from lxml import etree
    from typing_extensions import (
        NotRequired,
        Required,
        Self,
        TypeIs,
    )

    ElementT: TypeAlias = etree._Element
    DocumentT: TypeAlias = etree._ElementTree
    NodeT: TypeAlias = str | ElementT

    # In many places I treat lists as an "anonymous" element
    ElementishT: TypeAlias = ElementT | list[NodeT]

    # Can't actually do recursive types yet :(
    # Get as close as possible, but let lists be Any
    NodesT: TypeAlias = list[Any] | NodeT

    # Similar for JSON
    JSONT: TypeAlias = dict[str, Any]

    import sys
    from types import ModuleType

    if "Spec" not in sys.modules:
        from .Spec import Spec
    SpecT = Spec

    from .biblio import BiblioEntry
    from .config.BoolSet import BoolSet
    from .metadata import MetadataManager
    from .refs import MethodVariant, MethodVariants, ReferenceManager, RefSource, RefWrapper
    from .retrieve import DataFileRequester

    BiblioStorageT: TypeAlias = DefaultDict[str, list[BiblioEntry]]

    FillContainersT: TypeAlias = DefaultDict[str, list[ElementT]]

    LinkDefaultsT: TypeAlias = DefaultDict[str, list[tuple[str, str, str | None, str | None]]]



# pylint: disable=R1732

from __future__ import annotations

import io
import os

from . import InputSource, config, t
from . import messages as m

if t.TYPE_CHECKING:
    from .doctypes import Group, Org, Status


class DataFileRequester:
    def __init__(self, fileType: str | None = None, fallback: DataFileRequester | None = None) -> None:
        if fileType not in ("readonly", "latest"):
            msg = f"Bad value for DataFileRequester.type, got '{fileType}'."
            raise Exception(msg)
        self.fileType: str = fileType
        # fallback is another requester, used if the main one fails.
        self.fallback = fallback

    def path(self, *segs: str, fileType: str | None = None) -> str:
        return self._buildPath(segs=segs, fileType=fileType or self.fileType)

    @t.overload
    def fetch(self, *segs: str, str: t.Literal[True], okayToFail: bool = False, fileType: str | None = None) -> str: ...

    @t.overload
    def fetch(
        self,
        *segs: str,
        str: t.Literal[False],
        okayToFail: bool = False,
        fileType: str | None = None,
    ) -> t.TextIO: ...

    @t.overload
    def fetch(self, *segs: str, okayToFail: bool = False, fileType: str | None = None) -> t.TextIO: ...

    def fetch(
        self,
        *segs: str,
        str: bool = False,
        okayToFail: bool = False,
        fileType: str | None = None,
    ) -> str | t.TextIO:
        location = self._buildPath(segs=segs, fileType=fileType or self.fileType)
        try:
            if str:
                with open(location, encoding="utf-8") as fh:
                    return fh.read()
            else:
                return open(location, encoding="utf-8")
        except OSError:
            if self.fallback:
                try:
                    if str:
                        return self.fallback.fetch(*segs, str=True, okayToFail=okayToFail)
                    else:
                        return self.fallback.fetch(*segs, str=False, okayToFail=okayToFail)
                except OSError:
                    if str:
                        return self._fail(location, str=True, okayToFail=okayToFail)
                    else:
                        return self._fail(location, str=False, okayToFail=okayToFail)
            return self._fail(location, str, okayToFail)

    def walkFiles(self, *segs: str, fileType: str | None = None) -> t.Generator[str, None, None]:
        for _, _, files in os.walk(self._buildPath(segs, fileType=fileType or self.fileType)):
            yield from files

    def _buildPath(self, segs: t.Sequence[str], fileType: str | None = None) -> str:
        if fileType is None:
            fileType = self.fileType
        if fileType == "readonly":
            return config.scriptPath("spec-data", "readonly", *segs)
        else:
            return config.scriptPath("spec-data", *segs)

    @t.overload
    def _fail(self, location: str, str: t.Literal[True], okayToFail: bool) -> str: ...

    @t.overload
    def _fail(self, location: str, str: t.Literal[False], okayToFail: bool) -> t.TextIO: ...

    @t.overload
    def _fail(self, location: str, str: bool, okayToFail: bool) -> str | t.TextIO: ...

    def _fail(self, location: str, str: bool, okayToFail: bool) -> str | t.TextIO:
        if okayToFail:
            if str:
                return ""
            else:
                return io.StringIO("")
        msg = f"Couldn't find file '{location}'"
        raise OSError(msg)


defaultRequester = DataFileRequester(fileType="latest", fallback=DataFileRequester(fileType="readonly"))


def retrieveBoilerplateFile(
    doc: t.SpecT,
    name: str,
    group: Group | None = None,
    status: Status | None = None,
    org: Org | None = None,
    quiet: bool = False,
    allowLocal: bool = True,
    fileRequester: DataFileRequester | None = None,
) -> str:
    # Looks in three or four locations, in order:
    # the folder the spec source is in, the group's boilerplate folder, the megagroup's boilerplate folder, and the generic boilerplate folder.
    # In each location, it first looks for the file specialized on status, and then for the generic file.
    # Filenames must be of the format NAME.include or NAME-STATUS.include
    if fileRequester is None:
        dataFile = doc.dataFile
    else:
        dataFile = fileRequester

    if group is None:
        group = doc.doctype.group
    groupName = group.name.lower() if group else None
    if status is None:
        status = doc.doctype.status
    statusName = status.name.upper() if status else None
    if org is None:
        org = doc.doctype.org
    orgName = org.name.lower() if org else None

    searchLocally = allowLocal and doc.md.localBoilerplate[name]

    def boilerplatePath(*segs: str) -> str:
        return dataFile.path("boilerplate", *segs)

    filenames = []
    if statusName:
        filenames.append(f"{name}-{statusName}.include")
    filenames.append(f"{name}.include")

    sources: list[InputSource.InputSource] = []
    # 1: Look locally
    if searchLocally:
        for fn in filenames:
            source = doc.inputSource.relative(fn)
            if source:
                sources.append(source)
    else:
        for fn in filenames:
            if doc.inputSource.cheaplyExists(fn):
                m.warn(
                    f"Found {fn} next to the specification without a matching\n"
                    + f"Local Boilerplate: {name} yes\n"
                    + "in the metadata. This include won't be found when building via a URL.",
                )
                # We should remove this after giving specs time to react to the warning:
                source = doc.inputSource.relative(fn)
                if source:
                    sources.append(source)
    # 2: Look in the group's folder
    if groupName:
        sources.extend(InputSource.FileInputSource(boilerplatePath(groupName, fn), chroot=False) for fn in filenames)
    # 3: Look in the org's folder
    if orgName:
        sources.extend(InputSource.FileInputSource(boilerplatePath(orgName, fn), chroot=False) for fn in filenames)
    # 4: Look in the generic defaults
    sources.extend(InputSource.FileInputSource(boilerplatePath(fn), chroot=False) for fn in filenames)

    # Watch all the possible sources, not just the one that got used, because if
    # an earlier one appears, we want to rebuild.
    doc.recordDependencies(*sources)

    for source in sources:
        try:
            content = source.read().content
            return content
        except OSError:
            # That input doesn't exist.
            pass
    if not quiet:
        components = []
        if orgName:
            components.append(f"Org '{orgName}'")
        if groupName:
            components.append(f"Group '{groupName}'")
        if statusName:
            components.append(f"Status '{statusName}'")
        msg = "Couldn't find an appropriate include file for the {name} inclusion"
        if components:
            msg += ", given " + config.englishFromList(components, "and")
        else:
            msg += "."
        m.die(msg)
    return ""



from __future__ import annotations

from .. import config, t
from .. import messages as m

if t.TYPE_CHECKING:
    from . import DoctypeManager, Group, Org, Status  # pylint: disable=cyclic-import


def canonicalize(
    manager: DoctypeManager,
    rawOrg: str | None,
    rawGroup: str | None,
    rawStatus: str | None,
) -> tuple[Org | None, Group | None, Status | None]:
    # Takes raw Org/Status/Group names (something written in the Org/Status/Group metadata),
    # and, if possible, converts them into Org/Status/Group objects.

    # First, canonicalize the status and group casings, and separate them from
    # any inline org specifiers.
    # Then, figure out what the actual org name is.
    orgFromStatus, statusName = splitOrg(rawStatus)
    orgFromStatus = orgFromStatus.upper() if orgFromStatus is not None else None
    statusName = statusName.upper() if statusName is not None else None

    orgFromGroup, groupName = splitOrg(rawGroup)
    orgFromGroup = orgFromGroup.upper() if orgFromGroup is not None else None
    groupName = groupName.upper() if groupName is not None else None

    orgName = reconcileOrgs(rawOrg, orgFromStatus, orgFromGroup)

    if orgName is not None and rawOrg is None:
        orgInferredFrom = "Org (inferred from "
        if orgFromStatus is not None:
            orgInferredFrom += f"Status '{rawStatus}')"
        else:
            orgInferredFrom += f"Group '{rawGroup}')"
    else:
        orgInferredFrom = "Org"

    # Actually fetch the Org/Status/Group objects
    if orgName is not None:
        org = manager.getOrg(orgName)
        if org is None:
            m.die(f"Unknown {orgInferredFrom} '{orgName}'. See docs for recognized Org values.")
    else:
        org = None

    if groupName is not None:
        group = manager.getGroup(orgName, groupName)
        if group is None:
            if orgName is None:
                groups = manager.getGroups(orgName, groupName)
                if len(groups) > 1:
                    orgNamesForGroup = config.englishFromList((x.org.name for x in groups), "and")
                    m.die(
                        f"Your Group '{groupName}' exists under several Orgs ({orgNamesForGroup}). Specify which org you want in an Org metadata.",
                    )
                else:
                    m.die(f"Unknown Group '{groupName}'. See docs for recognized Group values.")
            else:
                groups = manager.getGroups(None, groupName)
                if len(groups) > 0:
                    orgNamesForGroup = config.englishFromList((x.org.name for x in groups), "and")
                    m.die(
                        f"Your Group '{groupName}' doesn't exist under the {orgInferredFrom} '{orgName}', but does exist under {orgNamesForGroup}. Specify the correct Org (or the correct Group).",
                    )
                else:
                    m.die(f"Unknown Group '{rawGroup}'. See docs for recognized Group values.")
    else:
        group = None

    # If Org wasn't specified anywhere, default it from Group if possible
    if org is None and group is not None:
        org = group.org

    if statusName is not None:
        if orgFromStatus is not None:
            # If status explicitly specified an org, use that
            status = manager.getStatus(orgFromStatus, statusName)
        elif org:
            # Otherwise, if we found an org, look for it there,
            # but fall back to looking for it in the generic statuses
            status = manager.getStatus(org.name, statusName, allowGeneric=True)
        else:
            # Otherwise, just look in the generic statuses;
            # the error stuff later will catch it if that doesn't work.
            status = manager.getStatus(None, statusName)
    else:
        # Just quick exit on this case, nothing we can do.
        return org, group, None

    # See if your org-specific Status matches your Org
    if org and status and status.org and status.org != org:
        m.die(f"Your {orgInferredFrom} is '{org.name}', but your Status is only usable in the '{status.org.name}' Org.")

    if group and status and status.org and status.org != group.org:
        # If using an org-specific Status, Group must match.
        # (Any group can use a generic status.)
        possibleStatusNames = [x.name for x in group.org.statuses.values()]
        m.die(
            f"Your Group ({group.name}) is in the '{group.org.name}' Org, but your Status ({status.name}) is only usable in the '{status.org.name}' Org. Allowed Status values for '{group.org.name}' are {config.englishFromList(sorted(possibleStatusNames))}",
        )

    if group and group.type is not None and status and status.groupTypes and group.type not in status.groupTypes:
        allowedStatuses = [s.name for s in group.org.statuses.values() if group.type in s.groupTypes]
        if allowedStatuses:
            m.warn(
                f"You used Status {status.name}, but your Group ({group.name}) is limited to the statuses {config.englishFromList(sorted(allowedStatuses))}.",
            )
        else:
            m.die(
                f"PROGRAMMING ERROR: Group '{group.fullName()}' has type '{group.type}', but none of the {group.org.name} Statuses are associated with that type.",
            )

    if group and status and group.org.name == "W3C":
        # Apply the special w3c rules
        validateW3CStatus(status)

    # Reconciliation done, return everything if Status exists.
    if status:
        return org, group, status

    # Otherwise, try and figure out why we failed to find the status

    possibleStatuses = manager.getStatuses(statusName)
    if len(possibleStatuses) == 0:
        m.die(f"Unknown Status metadata '{rawStatus}'. Check the docs for valid Status values.")
        return org, group, status
    elif len(possibleStatuses) == 1:
        possibleStatus = possibleStatuses[0]
        if possibleStatus.org is None:
            m.die(
                f"Your Status '{statusName}' is a generic status, but you explicitly specified '{rawStatus}'. Remove the org prefix from your Status.",
            )
        else:
            m.die(
                f"Your Status '{statusName}' only exists in the '{possibleStatus.org.name}' Org, but you specified the {orgInferredFrom} '{orgName}'.",
            )
    else:
        statusNames = config.englishFromList((x.org.name for x in possibleStatuses if x.org), "and")
        includesDefaultStatus = any(x.org is None for x in possibleStatuses)
        if includesDefaultStatus:
            msg = f"Your Status '{statusName}' only exists in Org(s) {statusNames}, or is a generic status."
        else:
            msg = f"Your Status '{statusName}' only exists in the Orgs {statusNames}."
        if orgName:
            if org:
                msg += f" Your specified {orgInferredFrom} is '{orgName}'."
            else:
                msg += f" Your specified {orgInferredFrom} is an unknown value '{orgName}'."
        else:
            msg += " Declare one of those Orgs in your Org metadata."
        m.die(msg)

    return org, group, status


def splitOrg(st: str | None) -> tuple[str | None, str | None]:
    if st is None:
        return None, None

    if "/" in st:
        parts = st.partition("/")
        return parts[0].strip().lower(), parts[2].strip()
    else:
        return None, st.strip()


def reconcileOrgs(fromRaw: str | None, fromStatus: str | None, fromGroup: str | None) -> str | None:
    # Since there are three potential sources of "org" name,
    # figure out what the name actually is,
    # and complain if they disagree.
    fromRaw = fromRaw.upper() if fromRaw else None
    fromStatus = fromStatus.upper() if fromStatus else None
    fromGroup = fromGroup.upper() if fromGroup else None

    orgName: str | None = fromRaw

    if fromStatus is not None:
        if orgName is None:
            orgName = fromStatus
        elif orgName == fromStatus:
            pass
        else:
            m.die(
                f"Your Org metadata specifies '{fromRaw}', but your Status metadata states an org of '{fromStatus}'. These must agree - either fix them or remove one of them.",
            )

    if fromGroup is not None:
        if orgName is None:
            orgName = fromGroup
        elif orgName == fromGroup:
            pass
        else:
            m.die(
                f"Your Org metadata specifies '{fromRaw}', but your Group metadata states an org of '{fromGroup}'. These must agree - either fix them or remove one of them.",
            )

    return orgName


def validateW3CStatus(status: Status) -> None:
    if status.name == "DREAM":
        m.warn("You used Status:DREAM for a W3C document. Consider Status:UD instead.")

    if status.name in ("IG-NOTE", "WG-NOTE"):
        m.die(
            f"Under Process2021, {status.name} is no longer a valid status. Use NOTE (or one of its variants NOTE-ED, NOTE-FPWD, NOTE-WD) instead.",
        )
    if status.name == "NOTE-FPWD":
        m.die(
            "Under Process2021, Notes no longer have a 'first public' draft stage. Use NOTE-WD instead.",
        )



from __future__ import annotations

import json
import os
import re
from collections import defaultdict

import requests

from .. import biblio, t
from .. import messages as m


def update(path: str, dryRun: bool = False) -> set[str] | None:
    m.say("Downloading biblio data...")
    biblios: t.BiblioStorageT = defaultdict(list)
    biblio.processSpecrefBiblioFile(getSpecrefData(), biblios, order=3)
    biblio.processSpecrefBiblioFile(getWG21Data(), biblios, order=3)
    writtenPaths: set[str] = set()
    if not dryRun:
        groupedBiblios, allNames = groupBiblios(biblios)
        # Save each group to a file
        for group, biblios in groupedBiblios.items():
            try:
                p = os.path.join(path, "biblio", f"biblio-{group}.data")
                writtenPaths.add(p)
                with open(p, "w", encoding="utf-8") as fh:
                    writeBiblioFile(fh, biblios)
            except Exception as e:
                m.die(f"Couldn't save biblio database to disk.\n{e}")
                return None

        # biblio-keys is used to help correct typos,
        # so remove all the useless purely-numbered name types,
        # as they aren't useful for typo-correction
        reducedNames = []
        for name in allNames:
            if re.search(r"-\d{8}$", name):
                continue
            if re.match(r"cwg\d+$", name):
                continue
            if re.match(r"ewg\d+$", name):
                continue
            if re.match(r"fs\d+$", name):
                continue
            if re.match(r"lewg\d+$", name):
                continue
            if re.match(r"lwg\d+$", name):
                continue
            if re.match(r"n\d+$", name):
                continue
            if re.match(r"p\d+(r\d+)?$", name):
                continue
            if re.match(r"d\d+(r\d+)?$", name):
                continue
            if re.match(r"rfc\d+$", name):
                continue
            if re.match(r"wg21-", name):
                continue
            if re.match(r"edit\d+$", name):
                continue
            if re.match(r"etsi-", name):
                continue
            if re.match(r"iso-\d+", name):
                continue
            if re.match(r"iso\d+", name):
                continue
            if re.match(r"scte\d+", name):
                continue
            reducedNames.append(name)
        try:
            p = os.path.join(path, "biblio-keys.json")
            writtenPaths.add(p)
            with open(p, "w", encoding="utf-8") as fh:
                fh.write(str(json.dumps(reducedNames, indent=0, ensure_ascii=False, sort_keys=True)))
        except Exception as e:
            m.die(f"Couldn't save biblio database to disk.\n{e}")
            return None

        # Collect all the number-suffix names which also exist un-numbered
        numberedNames = collectNumberedNames(reducedNames)
        try:
            p = os.path.join(path, "biblio-numeric-suffixes.json")
            writtenPaths.add(p)
            with open(p, "w", encoding="utf-8") as fh:
                fh.write(str(json.dumps(numberedNames, indent=0, ensure_ascii=False, sort_keys=True)))
        except Exception as e:
            m.die(f"Couldn't save biblio numeric-suffix information to disk.\n{e}")
    m.say("Success!")
    return writtenPaths


def getSpecrefData() -> str:
    try:
        return requests.get("https://api.specref.org/bibrefs", timeout=5).text
    except Exception as e:
        m.die(f"Couldn't download the SpecRef biblio data.\n{e}")
        return "{}"


def getWG21Data() -> str:
    try:
        return requests.get("https://wg21.link/specref.json", timeout=5).text
    except Exception as e:
        m.die(f"Couldn't download the WG21 biblio data.\n{e}")
        return "{}"


def getCSSWGData() -> list[str]:
    try:
        return requests.get(
            "https://raw.githubusercontent.com/w3c/csswg-drafts/main/biblio.ref",
            timeout=5,
        ).text.splitlines()
    except Exception as e:
        m.die(f"Couldn't download the CSSWG biblio data.\n{e}")
        return []


def groupBiblios(biblios: t.BiblioStorageT) -> tuple[dict[str, t.BiblioStorageT], list[str]]:
    # Group the biblios by the first two letters of their keys
    groupedBiblios: dict[str, t.BiblioStorageT] = {}
    allNames = []
    for k, v in sorted(biblios.items(), key=lambda x: x[0].lower()):
        allNames.append(k)
        group = k[0:2]
        if group not in groupedBiblios:
            groupedBiblios[group] = defaultdict(list)
        groupedBiblios[group][k] = v
    return groupedBiblios, allNames


def writeBiblioFile(fh: t.TextIO, biblios: t.BiblioStorageT) -> None:
    """
    Each line is a value for a specific key, in the order:

    key
    linkText
    date
    status
    title
    snapshot_url
    current_url
    obsoletedBy
    other
    etAl (as a boolish string)
    authors* (each on a separate line, an indeterminate number of lines)

    Each entry (including last) is ended by a line containing a single - character.
    """
    for key, entries in biblios.items():
        b = sorted(entries, key=lambda x: x.order)[0]
        if isinstance(b, biblio.NormalBiblioEntry):
            fh.write("d:" + key.lower() + "\n")
            fh.write(b.linkText + "\n")
            fh.write((b.date or "") + "\n")
            fh.write((b.status or "") + "\n")
            fh.write((b.title or "") + "\n")
            fh.write((b.snapshotURL or "") + "\n")
            fh.write((b.currentURL or "") + "\n")
            fh.write((b.obsoletedBy or "") + "\n")
            fh.write((b.other or "") + "\n")
            if b.etAl:
                fh.write("1\n")
            else:
                fh.write("\n")
            for author in b.authors:
                fh.write(author + "\n")
        elif isinstance(b, biblio.StringBiblioEntry):
            fh.write("s:" + key.lower() + "\n")
            fh.write(b.linkText + "\n")
            fh.write(b.data + "\n")
        elif isinstance(b, biblio.AliasBiblioEntry):
            fh.write("a:" + key.lower() + "\n")
            fh.write(b.linkText + "\n")
            fh.write(b.aliasOf + "\n")
        else:
            m.die(f"The biblio key '{key}' has an unknown biblio type '{format}'.")
            continue
        fh.write("-" + "\n")


def collectNumberedNames(names: t.Sequence[str]) -> defaultdict[str, list[str]]:
    """
    Collects the set of names that have numeric suffixes
    (excluding ones that look like dates)
    for better error-correction.
    """

    prefixes = defaultdict(list)
    for name in set(names):
        # Ignoring 4+ digits, as they're probably years or isodates.
        match = re.match(r"(.+\D)\d{1,3}$", name)
        if match:
            prefix = match.group(1)
            if prefix.endswith("-"):
                prefix = prefix[:-1]
            if prefix in names:
                prefixes[prefix].append(name)
                prefixes[prefix] = sorted(prefixes[prefix])
    return prefixes



from __future__ import annotations

from .. import h, t
from .. import messages as m


def requiredIDs(doc: t.SpecT) -> None:
    if not doc.md.requiredIDs:
        return
    doc_ids = {e.get("id") for e in h.findAll("[id]", doc)}
    for id in doc.md.requiredIDs:
        if id.startswith("#"):
            id = id[1:]
        if id not in doc_ids:
            m.die(f"Required ID '{id}' was not found in the document.")



from __future__ import annotations

import dataclasses
import re

from . import messages as m
from . import railroaddiagrams as rr
from . import t


def parse(string: str) -> rr.Diagram | None:
    """
    Parses a DSL for railroad diagrams,
    based on significant whitespace.
    Each command must be on its own line,
    and is written like "Sequence:\n".
    Children are indented on following lines.
    Some commands have non-child arguments;
    for block commands they are put on the same line, after the :,
    for text commands they are put on the same line *before* the :,
    like:
        Choice: 0
            Terminal: foo
            Terminal raw: bar
    """

    lines = string.splitlines()

    # Strip off any common initial whitespace from lines.
    initialIndent = t.cast(re.Match, re.match(r"(\s*)", lines[0])).group(1)
    for i, line in enumerate(lines):
        if line.startswith(initialIndent):
            lines[i] = line[len(initialIndent) :]
        else:
            m.die(f"Inconsistent indentation: line {i} is indented less than the first line.")
            return rr.Diagram()

    # Determine subsequent indentation
    for line in lines:
        match = re.match(r"(\s+)", line)
        if match:
            indentText = match.group(1)
            break
    else:
        indentText = "\t"

    # Turn lines into tree
    lastIndent = 0
    tree = RRCommand(name="Diagram", prelude="", children=[], text=None, line=0)
    activeCommands = {"0": tree}
    blockNames = "And|Seq|Sequence|Stack|Or|Choice|Opt|Optional|Plus|OneOrMore|Star|ZeroOrMore"
    textNames = "T|Terminal|N|NonTerminal|C|Comment|S|Skip"
    for i, line in enumerate(lines, 1):
        indent = 0
        while line.startswith(indentText):
            indent += 1
            line = line[len(indentText) :]
        if indent > lastIndent + 1:
            m.die(f"Line {i} jumps more than 1 indent level from the previous line:\n{line.strip()}")
            return rr.Diagram()
        lastIndent = indent
        if re.match(rf"\s*({blockNames})\W", line):
            match = re.match(r"\s*(\w+)\s*:\s*(.*)", line)
            if not match:
                m.die(f"Line {i} doesn't match the grammar 'Command: optional-prelude'. Got:\n{line.strip()}")
                return rr.Diagram()
            command = match.group(1)
            prelude = match.group(2).strip()
            node = RRCommand(name=command, prelude=prelude, children=[], text=None, line=i)
        elif re.match(rf"\s*({textNames})\W", line):
            match = re.match(r"\s*(\w+)(\s[\w\s]+)?:\s*(.*)", line)
            if not match:
                m.die(f"Line {i} doesn't match the grammar 'Command [optional prelude]: text'. Got:\n{line.strip()},")
                return rr.Diagram()
            command = match.group(1)
            if match.group(2):
                prelude = match.group(2).strip()
            else:
                prelude = None
            text = match.group(3).strip()
            node = RRCommand(
                name=command,
                prelude=prelude,
                children=[],
                text=text,
                line=i,
            )
        else:
            m.die(
                f"Line {i} doesn't contain a valid railroad-diagram command. Got:\n{line.strip()}",
            )
            return None

        activeCommands[str(indent)].children.append(node)
        activeCommands[str(indent + 1)] = node

    diagram = _createDiagram(tree)
    assert diagram is None or isinstance(diagram, rr.Diagram)
    return diagram


@dataclasses.dataclass
class RRCommand:
    name: str
    prelude: str | None
    children: list[RRCommand]
    text: str | None
    line: int


def _createDiagram(command: RRCommand) -> rr.DiagramItem | None:
    """
    From a tree of commands,
    create an actual Diagram class.
    Each command must be {command, prelude, children}
    """
    if command.name == "Diagram":
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        return rr.Diagram(*children)
    if command.name in ("T", "Terminal"):
        if command.children:
            m.die(f"Line {command.line} - Terminal commands cannot have children.")
            return None
        return rr.Terminal(command.text or "", command.prelude)
    if command.name in ("N", "NonTerminal"):
        if command.children:
            m.die(f"Line {command.line} - NonTerminal commands cannot have children.")
            return None
        return rr.NonTerminal(command.text or "", command.prelude)
    if command.name in ("C", "Comment"):
        if command.children:
            m.die(f"Line {command.line} - Comment commands cannot have children.")
            return None
        return rr.Comment(command.text or "", command.prelude)
    if command.name in ("S", "Skip"):
        if command.children:
            m.die(f"Line {command.line} - Skip commands cannot have children.")
            return None
        if command.text:
            m.die(f"Line {command.line} - Skip commands cannot have text.")
            return None
        return rr.Skip()
    if command.name in ("And", "Seq", "Sequence"):
        if command.prelude:
            m.die(f"Line {command.line} - Sequence commands cannot have preludes.")
            return None
        if not command.children:
            m.die(f"Line {command.line} - Sequence commands need at least one child.")
            return None
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        return rr.Sequence(*children)
    if command.name in ("Stack",):
        if command.prelude:
            m.die(f"Line {command.line} - Stack commands cannot have preludes.")
            return None
        if not command.children:
            m.die(f"Line {command.line} - Stack commands need at least one child.")
            return None
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        return rr.Stack(*children)
    if command.name in ("Or", "Choice"):
        if command.prelude == "":
            default = 0
        else:
            try:
                default = int(t.cast(str, command.prelude))
            except ValueError:
                m.die(f"Line {command.line} - Choice preludes must be an integer. Got:\n{command.prelude}")
                default = 0
        if not command.children:
            m.die(f"Line {command.line} - Choice commands need at least one child.")
            return None
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        return rr.Choice(default, *children)
    if command.name in ("Opt", "Optional"):
        if command.prelude not in (None, "", "skip"):
            m.die(f"Line {command.line} - Optional preludes must be nothing or 'skip'. Got:\n{command.prelude}")
            return None
        if len(command.children) != 1:
            m.die(f"Line {command.line} - Optional commands need exactly one child.")
            return None
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        return rr.Optional(children[0], skip=(command.prelude == "skip"))
    if command.name in ("Plus", "OneOrMore"):
        if command.prelude:
            m.die(f"Line {command.line} - OneOrMore commands cannot have preludes.")
            return None
        if 0 == len(command.children) > 2:
            m.die(f"Line {command.line} - OneOrMore commands must have one or two children.")
            return None
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        return rr.OneOrMore(*children)
    if command.name in ("Star", "ZeroOrMore"):
        if command.prelude not in (None, "", "skip"):
            m.die(f"Line {command.line} - ZeroOrMore preludes must be nothing or 'skip'. Got:\n{command.prelude}")
            return None
        if 0 == len(command.children) > 2:
            m.die(f"Line {command.line} - ZeroOrMore commands must have one or two children.")
            return None
        children = [_f for _f in [_createDiagram(child) for child in command.children] if _f]
        if not children:
            m.die(f"Line {command.line} - ZeroOrMore has no valid children.")
            return None
        repeat = children[1] if len(children) == 2 else None
        return rr.ZeroOrMore(children[0], repeat=repeat, skip=(command.prelude == "skip"))
    m.die(f"Line {command.line} - Unknown command '{command.name}'.")
    return None



from .external import ExternalRefsGroup, ExternalRefsManager, ExternalRefsSpec
from .manager import ReferenceManager
from .source import MethodVariant, MethodVariants, RefSource
from .wrapper import RefDataT, RefWrapper



from __future__ import annotations

from typing import get_args

import cddlparser

from . import config, h, t
from . import messages as m


class CDDLMarker(cddlparser.ast.Marker):
    """
    Marker that wraps CDDL definitions and references in <cddl> and <a> blocks
    so that cross-referencing logic may take place.
    """

    # Keep pointers on the current rule context to track generic parameters
    # that are scoped to it
    currentRule: cddlparser.ast.Rule | None
    currentParameters: list[str]

    # List of all CDDL terms defined so far to track and report duplicates
    defined: list

    def __init__(self) -> None:
        self.currentRule = None
        self.currentParameters = []
        self.defined = []

    def _recordDefinition(self, type: str, name: str, dfnFor: str | None = None) -> bool:
        for term in self.defined:
            if term["type"] == type and term["name"] == name and term["dfnFor"] == dfnFor:
                forText = "" if dfnFor is None else f' defined in type "{dfnFor}"'
                m.die(
                    f"CDDL {type} {name}{forText} creates a duplicate and cannot be referenced.\nPlease create additional CDDL types to disambiguate.",
                )
                return False
        if type != "parameter":
            for term in self.defined:
                if term["type"] != "parameter" and term["name"] == name and term["dfnFor"] == dfnFor:
                    forText = "" if dfnFor is None else f' defined in type "{dfnFor}"'
                    m.warn(
                        f"CDDL {type} {name}{forText} creates a duplicate with a CDDL {term['type']}.\nLink type needs to be specified to reference the term.\nConsider creating additional CDDL types to disambiguate.",
                    )
                    break
        term = {"type": type, "name": name, "dfnFor": dfnFor}
        self.defined.append(term)
        return True

    def serializeValue(self, prefix: str, value: str, suffix: str, node: cddlparser.ast.Value) -> str:
        name = prefix + value + suffix
        if node.type not in {"text", "bytes"}:
            return name
        parent = node.parentNode
        assert parent is not None
        if isinstance(parent, cddlparser.ast.Memberkey) and node.type == "text":
            # A literal text string also gives rise to a type
            # see RFC 8610, section 3.5.1:
            # https://datatracker.ietf.org/doc/html/rfc8610#section-3.5.1
            assert parent.parentNode is not None
            forName = self._getFor(parent.parentNode)
            if forName is None:
                # Cannot easily link member key back to a definition
                return name
            elif self._recordDefinition("key", value, forName):
                # Create a key with and without quotes as linking text
                lts = [value, name]
                return '<cddl data-cddl-type="key" data-cddl-for="{}" data-lt="{}">{}</cddl>'.format(
                    h.escapeAttr(forName),
                    h.escapeAttr("|".join(lts)),
                    name,
                )
            else:
                # Duplicate found, don't create a dfn
                return name
        elif isinstance(parent, cddlparser.ast.Operator) and parent.controller == node:
            # Probably a ".default" value. It may be possible to link the value
            # back to an enumeration but it's equally possible that this is just
            # a string that's not defined anywhere. Let's ignore.
            return name
        else:
            forName = self._getFor(node)
            if forName is None:
                return name
            elif self._recordDefinition("value", value, forName):
                lts = [value, name]
                return '<cddl data-cddl-type="value" data-cddl-for="{}" data-lt="{}">{}</cddl>'.format(
                    h.escapeAttr(forName),
                    h.escapeAttr("|".join(lts)),
                    name,
                )
            else:
                # Duplicate found, don't create a dfn
                return name

    def serializeName(self, name: str, node: cddlparser.ast.CDDLNode) -> str:
        # The node is a Typename. Such a node may appear in a Rule, a Type,
        # a Reference, a Memberkey, a GroupEntry, or GenericParameters
        parent = node.parentNode
        if isinstance(parent, cddlparser.ast.Rule):
            # Rule definition
            # Keep a pointer to the rule not to have to look for it again
            # when the function is called on the rule's children
            self.currentRule = parent
            self.currentParameters = []
            if parent.name.parameters is not None:
                assert isinstance(parent.name.parameters, cddlparser.ast.GenericParameters)
                self.currentParameters = [p.name for p in parent.name.parameters.parameters]
            if parent.assign.type in {cddlparser.Tokens.TCHOICEALT, cddlparser.Tokens.GCHOICEALT}:
                # The definition extends a base definition
                return '<a data-link-type="cddl-type" data-link-for="/">{}</a>'.format(name)
            elif self._recordDefinition("type", name):
                return '<cddl data-cddl-type="type" data-lt="{}">{}</cddl>'.format(h.escapeAttr(name), name)
            else:
                # Duplicate found, don't create a dfn
                return name
        elif isinstance(parent, cddlparser.ast.Memberkey):
            # Member definition
            if not parent.hasColon:
                # The key is actually a reference to a type
                if name in get_args(cddlparser.ast.PreludeType):
                    return '<a data-link-type="cddl-type" data-link-for="/" data-link-spec="rfc8610">{}</a>'.format(
                        name,
                    )
                else:
                    return '<a data-link-type="cddl-type" data-link-for="/">{}</a>'.format(name)
            assert parent.parentNode is not None
            forName = self._getFor(parent.parentNode)
            if forName is None:
                # Cannot easily link member key back to a definition
                return name
            else:
                lts = []
                if name[0] == '"':
                    lts = [name[1:-1], name]
                else:
                    lts = [name, '"' + name + '"']
                if self._recordDefinition("key", lts[0], forName):
                    return '<cddl data-cddl-type="key" data-cddl-for="{}" data-lt="{}">{}</cddl>'.format(
                        h.escapeAttr(forName),
                        h.escapeAttr("|".join(lts)),
                        name,
                    )
                else:
                    # Duplicate found, don't create a dfn
                    return name
        elif isinstance(parent, cddlparser.ast.GenericParameters):
            typename = parent.parentNode
            assert isinstance(typename, cddlparser.ast.Typename)
            if self._recordDefinition("parameter", name, typename.name):
                return '<cddl data-cddl-type="parameter" data-cddl-for="{}" data-lt="{}">{}</cddl>'.format(
                    h.escapeAttr(typename.name),
                    h.escapeAttr(name),
                    name,
                )
            else:
                # Duplicate found, don't create a dfn
                return name
        elif name in get_args(cddlparser.ast.PreludeType):
            # Link types that come from the CDDL prelude in RFC 8610
            return '<a data-link-type="cddl-type" data-link-for="/" data-link-spec="rfc8610">{}</a>'.format(name)
        elif name in self.currentParameters:
            # Name is a reference to a generic parameter
            assert self.currentRule is not None
            return '<a data-link-type="cddl-parameter" data-link-for="{}">{}</a>'.format(
                h.escapeAttr(self.currentRule.name.name),
                name,
            )
        else:
            return '<a data-link-type="cddl" data-link-for="/">{}</a>'.format(name)

    def _getFor(self, node: cddlparser.ast.CDDLNode) -> str | None:
        """
        Retrieve the "for" attribute for the node.
        """
        parent = node.parentNode
        while parent is not None:
            if isinstance(parent, cddlparser.ast.Rule):
                # Something defined in a rule
                return parent.name.name
            elif isinstance(parent, cddlparser.ast.GroupEntry) and parent.key is not None:
                # A type in a member key definition
                assert parent.parentNode is not None
                parentFor = self._getFor(parent.parentNode)
                if parentFor is None:
                    return parentFor
                if isinstance(parent.key.type, cddlparser.ast.Value) and parent.key.type.type == "text":
                    return parentFor + "/" + parent.key.type.value
                elif isinstance(parent.key.type, cddlparser.ast.Typename):
                    return parentFor + "/" + parent.key.type.name
                else:
                    return None
            parent = parent.parentNode
        return None


def markupCDDL(doc: t.SpecT) -> None:
    cddlEls = h.findAll("pre.cddl:not([data-no-cddl]), xmp.cddl:not([data-no-cddl])", doc)

    marker = CDDLMarker()
    for el in cddlEls:
        if h.isNormative(doc, el):
            text = h.textContent(el)
            try:
                ast = cddlparser.parse(text)
                h.replaceContents(el, h.parseHTML(ast.serialize(marker)))
            except Exception as err:
                m.die(
                    f"{err}\nInvalid CDDL block (first 100 characters):\n{text[0:100]}{'...' if len(text) > 100 else ''}",
                )
        h.addClass(doc, el, "highlight")
        doc.extraJC.addCDDLHighlighting()


def markupCDDLBlock(pre: t.ElementT, doc: t.SpecT) -> set[t.ElementT]:
    """
    Convert <cddl> blocks into "dfn" or links.
    """
    localDfns = set()

    for el in h.findAll("cddl", pre):
        # Prefix CDDL types with "cddl-" to avoid confusion with other
        # types (notably CSS ones such as "value")
        cddlType = "cddl-" + (el.get("data-cddl-type") or "")
        assert isinstance(cddlType, str)
        url = None
        ref = None
        cddlText: str
        for cddlText in (el.get("data-lt") or "").split("|"):
            linkFors: t.Sequence[str | None] | None = config.splitForValues(el.get("data-cddl-for"))
            if linkFors is None:
                linkFors = [None]
            for linkFor in linkFors:
                ref = doc.refs.getRef(
                    cddlType,
                    cddlText,
                    linkFor=linkFor,
                    status="local",
                    el=el,
                    error=True,
                )
                if isinstance(ref, str):
                    ref = None
                if ref:
                    url = ref.url
                    break
            if ref:
                break

        if url is None:
            el.tag = "dfn"
            el.set("data-dfn-type", cddlType)
            del el.attrib["data-cddl-type"]
            if el.get("data-cddl-for"):
                el.set("data-dfn-for", el.get("data-cddl-for") or "")
                del el.attrib["data-cddl-for"]
        else:
            # Copy over the auto-generated linking text to the manual dfn.
            # Note: "url" is not an absolute URL but rather a fragment ref. It
            # can thus be used as an ID selector to find the underlying dfn
            dfn = h.find(url, doc)
            assert dfn is not None
            lts = combineCDDLLinkingTexts(el.get("data-lt"), dfn.get("data-lt"))
            dfn.set("data-lt", lts)
            localDfns.add(dfn)

            # Reset the <cddl> element to be a link to the manual dfn.
            el.tag = "a"
            el.set("data-link-type", cddlType)
            el.set("data-lt", cddlText)
            del el.attrib["data-cddl-type"]
            if el.get("data-cddl-for"):
                el.set("data-link-for", el.get("data-cddl-for") or "")
                del el.attrib["data-cddl-for"]
            if el.get("id"):
                # ID was defensively added by the Marker.
                del el.attrib["id"]
    return localDfns


def combineCDDLLinkingTexts(t1: str | None, t2: str | None) -> str:
    t1s = (t1 or "").split("|")
    t2s = (t2 or "").split("|")
    for lt in t2s:
        if lt not in t1s:
            t1s.append(lt)
    return "|".join(t1s)



from __future__ import annotations

from . import h, t
from . import messages as m


def printjson(x: t.Any, indent: str | int = 2, level: int = 0) -> str:
    if isinstance(indent, int):
        # Convert into a number of spaces.
        indent = " " * indent
    x = getjson(x)
    if isinstance(x, dict):
        ret = printjsonobject(x, indent, level)
    elif isinstance(x, list):
        if len(x) > 0 and isinstance(getjson(x[0]), dict):
            ret = printjsonobjectarray(x, indent, level)
        else:
            ret = printjsonsimplearray(x, indent, level)
    else:
        ret = printjsonprimitive(x)
    if level == 0 and ret.startswith("\n"):
        ret = ret[1:]
    return ret
    # return json.dumps(obj, indent=2, default=lambda x:x.__json__())


def getjson(x: t.Any) -> t.Any:
    try:
        return x.__json__()
    except AttributeError:
        return x


def printjsonobject(x: t.Any, indent: str, level: int) -> str:
    x = getjson(x)
    ret = ""
    maxKeyLength = 0
    for k in x:
        maxKeyLength = max(maxKeyLength, len(k))
    for k, v in x.items():
        ret += (
            "\n"
            + (indent * level)
            + m.printColor((k + ": ").ljust(maxKeyLength + 2), "cyan")
            + printjson(v, indent, level + 1)
        )
    return ret


def printjsonobjectarray(x: list[dict[str, t.Any]], indent: str, level: int) -> str:
    # Prints an array of objects
    x = getjson(x)
    ret = ""
    for i, v in enumerate(x):
        if i != 0:
            ret += "\n" + (indent * level) + m.printColor("=" * 10, "blue")
        ret += printjsonobject(v, indent, level)
    return ret


def printjsonsimplearray(x: list, indent: str, level: int) -> str:  # pylint: disable=unused-argument
    x = getjson(x)
    ret = m.printColor("[", "blue")
    for i, v in enumerate(x):
        if i != 0:
            ret += m.printColor(", ", "blue")
        ret += printjsonprimitive(v)
    ret += m.printColor("]", "blue")
    return ret


def printjsonprimitive(x: t.Any) -> str:
    x = getjson(x)
    if isinstance(x, int):
        return str(x)
    if isinstance(x, str):
        return x
    if isinstance(x, bool):
        return str(x)
    if x is None:
        return "null"
    if h.isElement(x):
        return repr(x) + ":" + h.outerHTML(x)
    msg = f"Could not print value: {x}"
    raise ValueError(msg)



from __future__ import annotations

import copy
import dataclasses
import re
from collections import defaultdict

from .. import config, constants, retrieve, t
from .. import messages as m
from . import utils, wrapper
from .utils import LinkFailure

LAZY_LOADED_SOURCES: list[str] = ["foreign"]


class RefSource:
    __slots__ = [
        "_loadedAnchorGroups",
        "dataFile",
        "fors",
        "ignoredSpecs",
        "methods",
        "refs",
        "replacedSpecs",
        "source",
        "specs",
    ]

    # Which sources use lazy-loading; other sources always have all their refs loaded immediately.

    def __init__(
        self,
        source: str,
        specs: dict[str, dict[str, str]] | None = None,
        ignored: set[str] | None = None,
        replaced: set[tuple[str, str]] | None = None,
        fileRequester: t.DataFileRequester | None = None,
    ) -> None:
        if fileRequester is None:
            self.dataFile = retrieve.defaultRequester
        else:
            self.dataFile = fileRequester

        # String identifying which refsource this is.
        self.source = source

        # Dict of {linking text => [anchor data]}
        self.refs: defaultdict[str, list[t.RefWrapper]] = defaultdict(list)

        # Dict of {argless method signatures => {"argfull signature": {"args":[args], "for":[fors]}}}
        self.methods: dict[str, MethodVariants] = {}

        # Dict of {for value => [terms]}
        self.fors: defaultdict[str, list[str]] = defaultdict(list)

        self.specs: dict[str, dict[str, str]] = {} if specs is None else specs
        self.ignoredSpecs = set() if ignored is None else ignored
        self.replacedSpecs = set() if replaced is None else replaced
        self._loadedAnchorGroups: set[str] = set()

    def fetchRefs(self, key: str) -> list[t.RefWrapper]:
        """Safe, lazy-loading version of self.refs[key]"""

        if key in self.refs:
            return self.refs[key]

        if self.source not in LAZY_LOADED_SOURCES:
            return []

        group = config.groupFromKey(key)
        if group in self._loadedAnchorGroups:
            # Group was loaded, but previous check didn't find it, so it's just not here.
            return []
        # Otherwise, load the group file.
        with self.dataFile.fetch("anchors", f"anchors-{group}.data", okayToFail=True) as fh:
            self.refs.update(decodeAnchors(fh))
            self._loadedAnchorGroups.add(group)
        return self.refs.get(key, [])

    def fetchAllRefs(self) -> list[tuple[str, list[t.RefWrapper]]]:
        """Nuts to lazy-loading, just load everything at once."""

        if self.source not in LAZY_LOADED_SOURCES:
            return list(self.refs.items())

        for file in self.dataFile.walkFiles("anchors"):
            group = t.cast(re.Match, re.match(r"anchors-(.{2})", file)).group(1)
            if group in self._loadedAnchorGroups:
                # Already loaded
                continue
            with self.dataFile.fetch("anchors", file) as fh:
                self.refs.update(decodeAnchors(fh))
                self._loadedAnchorGroups.add(group)
        return list(self.refs.items())

    def queryRefs(
        self,
        text: str | None = None,
        spec: str | None = None,
        linkType: str | None = None,
        linkFor: str | list[str] | None = None,
        explicitFor: bool = False,
        linkForHint: str | None = None,
        status: str | None = None,
        statusHint: str | None = None,
        export: bool | None = None,
        ignoreObsoletes: bool = False,
        latestOnly: bool = True,
        dedupURLs: bool = True,
        exact: bool | None = None,
        error: bool = False,
        el: t.ElementT | None = None,
    ) -> tuple[list[t.RefWrapper], LinkFailure | None, list[t.RefWrapper]]:
        if exact is not None:
            return self._queryRefs(
                text=text,
                spec=spec,
                linkType=linkType,
                linkFor=linkFor,
                explicitFor=explicitFor,
                linkForHint=linkForHint,
                status=status,
                statusHint=statusHint,
                export=export,
                ignoreObsoletes=ignoreObsoletes,
                latestOnly=latestOnly,
                dedupURLs=dedupURLs,
                error=error,
                el=el,
                exact=exact,
            )
        else:
            # First search for the exact term, and only if it fails fall back to conjugating.
            results, linkFailure, _ = self._queryRefs(
                text=text,
                spec=spec,
                linkType=linkType,
                linkFor=linkFor,
                explicitFor=explicitFor,
                linkForHint=linkForHint,
                status=status,
                statusHint=statusHint,
                export=export,
                ignoreObsoletes=ignoreObsoletes,
                latestOnly=latestOnly,
                dedupURLs=dedupURLs,
                error=error,
                el=el,
                exact=True,
            )
            if linkFailure:
                return self._queryRefs(
                    text=text,
                    spec=spec,
                    linkType=linkType,
                    linkFor=linkFor,
                    explicitFor=explicitFor,
                    linkForHint=linkForHint,
                    status=status,
                    statusHint=statusHint,
                    export=export,
                    ignoreObsoletes=ignoreObsoletes,
                    latestOnly=latestOnly,
                    dedupURLs=dedupURLs,
                    error=error,
                    el=el,
                    exact=False,
                )
            else:
                return results, None, []

    def allRefsIterator(self) -> t.Generator[t.RefWrapper, None, None]:
        # Turns a dict of arrays of refs into an iterator of refs
        for _, group in self.fetchAllRefs():
            yield from group

    def textRefsIterator(self, texts: list[str]) -> t.Generator[t.RefWrapper, None, None]:
        # Same as above, but only grabs those keyed to a given text
        for x in texts:
            yield from self.fetchRefs(x)

    def forRefsIterator(self, targetFors: str | list[str]) -> t.Generator[t.RefWrapper, None, None]:
        # Same as above, but only grabs those for certain values
        if isinstance(targetFors, str):
            targetFors = [targetFors]
        for for_ in targetFors:
            for x in self.fors[for_]:
                yield from self.fetchRefs(x)

    def _queryRefs(
        self,
        text: str | None = None,
        spec: str | None = None,
        linkType: str | None = None,
        linkFor: str | list[str] | None = None,
        explicitFor: bool = False,
        linkForHint: str | None = None,
        status: str | None = None,
        statusHint: str | None = None,
        export: bool | None = None,
        ignoreObsoletes: bool = False,
        latestOnly: bool = True,
        dedupURLs: bool = True,
        exact: bool = False,
        error: bool = False,
        el: t.ElementT | None = None,
    ) -> tuple[list[t.RefWrapper], LinkFailure | None, list[t.RefWrapper]]:
        # Query the ref database.
        # If it fails to find a ref, also returns the stage at which it finally ran out of possibilities.

        # Set up the initial list of refs to query
        if text:
            if exact:
                refs = list(self.textRefsIterator([text]))
            else:
                textsToSearch = list(utils.linkTextVariations(text, linkType))

                if any(st.endswith("()") for st in textsToSearch):
                    # Let argless methods (either with () at the end, or no parens at all)
                    # pull in their argful variants for searching.
                    for st in textsToSearch[:]:
                        textsToSearch += getArgfulMethodVariants(st, self)

                if (linkType is None or linkType in config.lowercaseTypes) and text.lower() != text:
                    textsToSearch += [t.lower() for t in textsToSearch]
                refs = list(self.textRefsIterator(textsToSearch))
        elif linkFor:
            refs = list(self.forRefsIterator(linkFor))
        else:
            refs = list(self.allRefsIterator())
        if not refs:
            return [], LinkFailure.Text, []

        oldRefs = refs
        if linkType:
            if linkType in config.dfnTypes:
                linkTypes = [linkType]
            elif linkType in config.linkTypeToDfnType:
                linkTypes = list(config.linkTypeToDfnType[linkType])
            else:
                if error:
                    m.linkerror(f"Unknown link type '{linkType}'.", el=el)
                return [], LinkFailure.Type, []
            refs = [x for x in refs if x.type in linkTypes]
        if not refs:
            return [], LinkFailure.Type, oldRefs

        oldRefs = refs
        if export is not None:
            refs = [x for x in refs if x.export == export]
        if not refs:
            return [], LinkFailure.Export, oldRefs

        oldRefs = refs
        if spec:
            refs = [x for x in refs if spec in (x.spec, x.shortname)]
        if not refs:
            return [], LinkFailure.Spec, oldRefs

        """
        for=A | forHint=B | explicitFor
        ✘ | ✘ | ✘ = anything
        ✘ | ✘ | ✔ = /
        ✘ | ✔ | ✘ = B/, fallback to anything
        ✘ | ✔ | ✔ = B/, fallback to /
        ✔ | ✘ | ✘ = A/
        ✔ | ✘ | ✔ = A/
        ✔ | ✔ | ✘ = A/
        ✔ | ✔ | ✔ = A/
        """

        oldRefs = refs
        if linkFor:
            refs = filterByFor(refs, linkFor)
        elif linkForHint:
            if explicitFor:
                tempRefs = filterByFor(refs, linkForHint)
                if not tempRefs:
                    tempRefs = filterByFor(refs, "/")
                refs = tempRefs
            else:
                # Handled later, in the "just a hint" section.
                pass
        elif explicitFor:
            refs = filterByFor(refs, "/")
        if not refs:
            return [], LinkFailure.For, oldRefs

        oldRefs = refs
        if status:
            refs = filterByStatus(self, refs, status)
        if not refs:
            return [], LinkFailure.Status, refs

        if status is None and statusHint:
            tempRefs = filterByStatus(self, refs, statusHint)
            if tempRefs:
                refs = tempRefs

        if linkForHint and not linkFor and not explicitFor:
            tempRefs = filterByFor(refs, linkForHint)
            if tempRefs:
                refs = tempRefs

        oldRefs = refs
        if ignoreObsoletes and not spec:
            # Remove any ignored or obsoleted specs
            # If you specified the spec, don't filter things - you know what you're doing.
            refs = utils.filterObsoletes(refs, replacedSpecs=self.replacedSpecs, ignoredSpecs=self.ignoredSpecs)
        if not refs:
            return [], LinkFailure.IgnoredSpecs, oldRefs

        if dedupURLs:
            # With non-exact texts, you might have multiple "anchors"
            # that point to the same url. Dedup them.
            seenUrls = set()
            tempRefs = []
            # Sort the refs so the kept one doesn't depend on ordering in the RefSource dict.
            for ref in sorted(copy.copy(refs), key=lambda x: x.text):
                if ref.url not in seenUrls:
                    tempRefs.append(ref)
                    seenUrls.add(ref.url)
            refs = tempRefs

        if latestOnly:
            # If multiple levels of the same shortname exist,
            # only use the latest level.
            # If generating for a snapshot, prefer the latest snapshot level,
            # unless that doesn't exist, in which case just prefer the latest level.
            refs = utils.filterOldVersions(refs, status=status or statusHint)

        return refs, None, []

    def addMethodVariants(self, methodSig: str, forVals: t.Iterable[str], shortname: str | None) -> None:
        # Takes a full method signature, like "foo(bar)",
        # and adds appropriate lines to self.methods for it
        match = re.match(r"([^(]+)\((.*)\)", methodSig)
        if not match:
            # Was fed something that's not a method signature.
            return None
        name, args = match.groups()
        arglessMethodSig = name + "()"
        if arglessMethodSig not in self.methods:
            self.methods[arglessMethodSig] = MethodVariants(arglessMethodSig, {})
        variants = self.methods[arglessMethodSig].variants
        if methodSig not in variants:
            args = [x.strip() for x in args.split(",")]
            variants[methodSig] = MethodVariant(signature=methodSig, args=args, for_=[], shortname=shortname)
        variants[methodSig].for_.extend(forVals)


def filterByFor(refs: t.Sequence[t.RefWrapper], linkFor: str | list[str]) -> list[t.RefWrapper]:
    return [x for x in refs if matchFor(x.for_, linkFor)]


def matchFor(forVals: list[str], forTest: str | list[str]) -> bool:
    # forTest can be a string, either / for no for or the for value to match,
    # or an array of strings, of which any can match
    if forTest == "/":
        return not bool(forVals)
    if isinstance(forTest, str):
        return forTest in forVals
    return any(matchFor(forVals, ft) for ft in forTest)


def filterByStatus(source: RefSource, refs: t.Sequence[t.RefWrapper], status: str) -> list[t.RefWrapper]:
    if status in constants.refStatus:
        # If status is "current'", kill snapshot refs unless their spec *only* has a snapshot_url
        if status == constants.refStatus.current:
            return [
                ref
                for ref in refs
                if ref.status == "current"
                or (ref.status == "snapshot" and source.specs.get(ref.spec, {}).get("current_url") is None)
            ]
        # If status is "snapshot", kill current refs if there's a corresponding snapshot ref for the same spec.
        elif status == constants.refStatus.snapshot:
            snapshotSpecs = [ref.spec for ref in refs if ref.status == "snapshot"]
            return [
                ref
                for ref in refs
                if ref.status == "snapshot" or (ref.status == "current" and ref.spec not in snapshotSpecs)
            ]
    # Status is a non-refStatus, but is a valid linkStatus, like "local"
    elif status in config.linkStatuses:
        return [x for x in refs if x.status == status]

    msg = f"Status value of '{status}' not handled"
    raise Exception(msg)


def decodeAnchors(linesIter: t.Iterator[str]) -> defaultdict[str, list[t.RefWrapper]]:
    # Decodes the anchor storage format into {key: [{anchor-data}]}
    anchors = defaultdict(list)
    try:
        while True:
            aText = next(linesIter)[:-1]
            displayText = next(linesIter)[:-1]
            data: wrapper.RefDataT = {
                "type": next(linesIter),
                "spec": next(linesIter),
                "shortname": next(linesIter),
                "level": next(linesIter),
                "status": next(linesIter),
                "url": next(linesIter),
                "export": next(linesIter) != "\n",
                "normative": next(linesIter) != "\n",
                "for_": [],
            }
            while True:
                line = next(linesIter)
                if line == "-\n":
                    break
                t.cast("list", data["for_"]).append(line)
            anchors[aText].append(wrapper.RefWrapper(aText, displayText, data))
    except StopIteration:
        return anchors


def getArgfulMethodVariants(maybeMethodSig: str, refs: RefSource) -> list[str]:
    if maybeMethodSig.endswith("()") and maybeMethodSig in refs.methods:
        return list(refs.methods[maybeMethodSig].variants.keys())
    return []


@dataclasses.dataclass
class MethodVariants:
    arglessSignature: str
    variants: dict[str, MethodVariant]


@dataclasses.dataclass
class MethodVariant:
    signature: str
    args: list[str]
    for_: list[str]
    shortname: str | None



from __future__ import annotations

import io
import itertools

from .. import t
from . import dom

if t.TYPE_CHECKING:
    WriterFn: t.TypeAlias = t.Callable[[str], t.Any]

    # more specific than t.NodesT, as nested lists can't happen
    Nodes: t.TypeAlias = t.ElementT | list[str | t.ElementT]
    Blocks: t.TypeAlias = list[Nodes]


class Serializer:
    inlineEls = frozenset(
        [
            "a",
            "em",
            "strong",
            "small",
            "s",
            "cite",
            "q",
            "dfn",
            "abbr",
            "data",
            "time",
            "code",
            "var",
            "samp",
            "kbd",
            "sub",
            "sup",
            "i",
            "b",
            "u",
            "mark",
            "ruby",
            "bdi",
            "bdo",
            "span",
            "br",
            "wbr",
            "img",
            "meter",
            "progress",
            "math",
            "[]",
        ],
    )
    rawEls = frozenset(["xmp", "script", "style"])
    voidEls = frozenset(
        [
            "area",
            "base",
            "br",
            "col",
            "command",
            "embed",
            "hr",
            "img",
            "input",
            "keygen",
            "link",
            "meta",
            "param",
            "source",
            "track",
            "wbr",
        ],
    )
    omitEndTagEls = frozenset(
        [
            "td",
            "th",
            "tr",
            "thead",
            "tbody",
            "tfoot",
            "colgroup",
            "col",
            "li",
            "dt",
            "dd",
            "html",
            "head",
            "body",
        ],
    )

    def __init__(self, opaqueElements: t.Iterable[str], blockElements: t.Iterable[str]) -> None:
        self.opaqueEls = frozenset(opaqueElements)
        self.blockEls = frozenset(blockElements)

    def serialize(self, tree: t.DocumentT) -> str:
        output = io.StringIO()
        writer = output.write
        writer("<!doctype html>")
        self._serializeEl(tree.getroot(), writer)
        s = output.getvalue()
        output.close()
        return s

    def unfuckName(self, n: str) -> str:
        # LXML does namespaces stupidly
        if n.startswith("{"):
            return n.partition("}")[2]
        return n

    def groupIntoBlocks(self, nodes: t.Iterable[t.NodeT]) -> Blocks:
        # Turns the iterable into a list of block elements and inline "spans"
        # (a list of text and inline elements).
        blocks: Blocks = []
        for node in nodes:
            if self.isElement(node) and self.isBlockElement(node.tag):
                blocks.append(node)
            else:
                if len(blocks) == 0 or not isinstance(blocks[-1], list):
                    blocks.append([])
                lastBlock = t.cast("list[str | t.ElementT]", blocks[-1])
                lastBlock.append(node)
        return blocks

    def startTag(self, tag: str, el: Nodes, write: WriterFn) -> None:
        if isinstance(el, list):
            return
        if not dom.hasAttrs(el):
            write("<" + tag + ">")
            return

        strs = []
        strs.append("<" + tag)
        for attrName, attrVal in sorted(el.items()):
            if str(attrName).startswith("bs-"):
                # Skip bs- prefixed attributes, as they're used
                # for Bikeshed-internal purposes.
                continue
            if attrVal == "":
                strs.append(" " + self.unfuckName(str(attrName)))
            elif attrName == "class" and " " in attrVal:
                sortedClasses = " ".join(sorted([x for x in str(attrVal).split(r" ") if x]))
                strs.append(f' class="{sortedClasses}"')
            else:
                strs.append(" " + self.unfuckName(str(attrName)) + '="' + dom.escapeAttr(str(attrVal)) + '"')
        strs.append(">")
        write("".join(strs))

    def endTag(self, tag: str, write: WriterFn) -> None:
        if tag != "[]":
            write("</" + tag + ">")

    def isElement(self, node: t.Any) -> t.TypeGuard[t.ElementT]:
        return dom.isElement(node)

    def isAnonBlock(self, block: t.Any) -> bool:
        return not dom.isElement(block)

    def isVoidElement(self, tag: str) -> bool:
        return tag in self.voidEls

    def isRawElement(self, tag: str) -> bool:
        return tag in self.rawEls

    def isOpaqueElement(self, tag: str) -> bool:
        return tag in self.opaqueEls

    def isInlineElement(self, tag: str) -> bool:
        return (tag in self.inlineEls) or ("-" in tag and tag not in self.blockEls)

    def isBlockElement(self, tag: str) -> bool:
        return not self.isInlineElement(tag)

    def needsEndTag(self, el: t.ElementT, nextEl: Nodes | None = None) -> bool:
        if el.tag not in self.omitEndTagEls:
            return True
        if el.tag in ["dt", "dd"]:
            if nextEl is None:
                return False
            if self.isElement(nextEl) and nextEl.tag in ["dt", "dd"]:  # noqa: SIM103
                return False
            return True
        return False

    def justWS(self, block: t.NodesT) -> bool:
        if self.isElement(block):
            return False
        return len(block) == 1 and isinstance(block[0], str) and block[0].strip() == ""

    def _writeVoidElement(self, tag: str, el: t.ElementT, write: WriterFn, indent: int) -> None:
        write(" " * indent)
        self.startTag(tag, el, write)

    def _writeRawElement(self, tag: str, el: t.ElementT, write: WriterFn, indent: int) -> None:
        if tag == "script" and el.get("src") is not None:
            # A *linking* script, doesn't need to be treated specially.
            write(" " * indent)
            self.startTag(tag, el, write)
            self.endTag(tag, write)
            return

        # Otherwise, dedent completely, since I don't know if indenting
        # is significant for the element.
        self.startTag(tag, el, write)
        for node in dom.childNodes(el, mergeText=True):
            if self.isElement(node):
                msg = f"Somehow a CDATA element got an element child:\n{dom.outerHTML(el)}"
                raise Exception(msg)
            else:
                assert isinstance(node, str)
                write(node)
        self.endTag(tag, write)

    def _writeOpaqueElement(self, tag: str, el: t.ElementT, write: WriterFn, indent: int) -> None:
        self.startTag(tag, el, write)
        for node in dom.childNodes(el, mergeText=True):
            if self.isElement(node):
                self._serializeEl(node, write, indent=indent, pre=True)
            else:
                assert isinstance(node, str)
                write(dom.escapeHTML(node))
        self.endTag(tag, write)

    def _writeInlineElement(self, tag: str, el: Nodes, write: WriterFn, inline: bool) -> None:
        self.startTag(tag, el, write)
        for node in dom.childNodes(el, mergeText=True):
            if self.isElement(node):
                self._serializeEl(node, write, inline=inline)
            else:
                assert isinstance(node, str)
                write(dom.escapeHTML(node))
        self.endTag(tag, write)

    def _blocksFromChildren(self, children: t.Iterable[t.NodeT]) -> Nodes:
        return t.cast("Nodes", [block for block in self.groupIntoBlocks(children) if not self.justWS(block)])

    def _categorizeBlockChildren(self, el: Nodes) -> tuple[str, Nodes | None]:
        """
        Figure out what sort of contents the block has,
        so we know what serialization strategy to use.
        """

        # Early exit for a totally empty element (including just ws)
        # This will fail if there are odd nodes, even tho they're not serialized.
        if self.isElement(el) and len(el) == 0 and dom.emptyText(el.text):
            return "empty", None

        # See if there are any block children
        children = dom.childNodes(el, clear=True, mergeText=True)
        for child in children:
            if self.isElement(child) and self.isBlockElement(child.tag):
                return "blocks", self._blocksFromChildren(children)

        # See if it's an empty element *when ignoring odd nodes*,
        # which childNodes() does by default
        if all(isinstance(x, str) and dom.emptyText(x) for x in children):
            return "empty", None

        # Otherwise, it's all inline content.
        return "inlines", children

    def _writeBlockElement(self, tag: str, el: t.ElementT, write: WriterFn, indent: int, nextEl: Nodes | None) -> None:
        # Dropping pure-WS anonymous blocks.
        # This maintains whitespace between *inline* elements, which is required.
        # It just avoids serializing a line of "inline content" that's just WS.
        contentsType, contents = self._categorizeBlockChildren(el)

        if contentsType == "empty":
            # Empty of text and children
            assert self.isElement(el)
            write(" " * indent)
            self.startTag(tag, el, write)
            if self.needsEndTag(el, nextEl):
                self.endTag(tag, write)
        elif contentsType == "inlines":
            # Contains only inlines, print accordingly
            assert contents is not None
            # el might be a list of inline content
            write(" " * indent)
            self.startTag(tag, el, write)
            self._serializeEl(contents, write, inline=True)
            if self.needsEndTag(el, nextEl):
                self.endTag(tag, write)
            return
        else:
            # Otherwise I'm a block that contains at least one block
            assert contents is not None
            assert self.isElement(el)
            write(" " * indent)
            self.startTag(tag, el, write)
            for block, nextBlock in t.cast("itertools.zip_longest[tuple[Nodes, Nodes|None]]", pairwise(contents)):
                if isinstance(block, list):
                    # is an array of inlines
                    if len(block) > 0:
                        write("\n" + (" " * (indent + 1)))
                        self._serializeEl(block, write, inline=True)
                else:
                    write("\n")
                    self._serializeEl(block, write, indent=indent + 1, nextEl=nextBlock)
            if self.needsEndTag(el, nextEl):
                write("\n" + (" " * indent))
                self.endTag(tag, write)

    def _serializeEl(
        self,
        el: Nodes,
        write: WriterFn,
        indent: int = 0,
        pre: bool = False,
        inline: bool = False,
        nextEl: Nodes | None = None,
    ) -> None:
        if isinstance(el, list):
            tag = "[]"
        else:
            tag = self.unfuckName(el.tag)

        if self.isVoidElement(tag):
            assert self.isElement(el)
            self._writeVoidElement(tag, el, write, indent)
        elif self.isRawElement(tag):
            assert self.isElement(el)
            self._writeRawElement(tag, el, write, indent)
        elif pre or self.isOpaqueElement(tag):
            assert self.isElement(el)
            self._writeOpaqueElement(tag, el, write, indent)
        elif inline:
            self._writeInlineElement(tag, el, write, inline)
        else:
            assert self.isElement(el)
            self._writeBlockElement(tag, el, write, indent, nextEl)


if t.TYPE_CHECKING:
    PairwiseU = t.TypeVar("PairwiseU")


def pairwise(iterable: t.Iterable[PairwiseU]) -> itertools.zip_longest[tuple[PairwiseU, PairwiseU | None]]:
    # pairwise('ABCDEFG') --> AB BC CD DE EF FG GNone
    a, b = itertools.tee(iterable)
    next(b, None)
    return itertools.zip_longest(a, b)



from __future__ import annotations

import widlparser  # pylint: disable=unused-import

from .. import messages as m
from .. import t


def missingExposed(doc: t.SpecT) -> None:
    """
    Checks that every IDL interface or namespace has an [Exposed] attribute.
    Specifically:
    * Any namespace
    * Any non-callback interface without [NoInterfaceObject]
    * Any callback interface that has constants in it.

    Partials never need to declare [Exposed], I think?
    """
    if not doc.md.complainAbout["missing-exposed"]:
        return

    for construct in doc.widl.constructs:
        extendedAttrs: list[str]
        if construct.extended_attributes is None:
            extendedAttrs = []
        else:
            extendedAttrs = [x.name for x in construct.extended_attributes if x.name is not None]
        if construct.idl_type == "namespace":
            good = False
            for attr in extendedAttrs:
                if attr == "Exposed":
                    good = True
                    break
            if not good:
                m.lint(
                    f"The '{construct.name}' namespace is missing an [Exposed] extended attribute. Does it need [Exposed=Window], or something more?",
                )
        elif construct.idl_type == "interface":
            good = False
            for attr in extendedAttrs:
                if attr == "Exposed":
                    good = True
                    break
                if attr == "NoInterfaceObject":
                    good = True
                    break
            if not good:
                m.lint(
                    f"The '{construct.name}' interface is missing an [Exposed] extended attribute. Does it need [Exposed=Window], or something more?",
                )
        elif construct.idl_type == "callback":
            if not hasattr(construct, "interface"):
                # Just a callback function, it's fine
                continue



from __future__ import annotations

from . import constants, retrieve, t


def load(doc: t.SpecT) -> None:
    code = retrieve.retrieveBoilerplateFile(doc, "bs-extensions", allowLocal=constants.executeCode)
    exec(code, globals())



from .BoolSet import BoolSet
from .dfnTypes import (
    adjustKey,
    anchorishElements,
    cddlTypes,
    cssTypes,
    dfnClassToType,
    dfnElements,
    dfnElementsSelector,
    dfnTypes,
    functionishTypes,
    idlMethodTypes,
    idlNameTypes,
    idlTypes,
    linkStatuses,
    linkTypeIn,
    linkTypes,
    linkTypeToDfnType,
    lowercaseTypes,
    markupTypes,
    maybeTypes,
    specStatuses,
    typeRe,
    typesNotUsingFor,
    typesUsingFor,
)
from .main import (
    chrootPath,
    docPath,
    doEvery,
    englishFromList,
    flatten,
    groupFromKey,
    intersperse,
    processTextNodes,
    reSubObject,
    safeIndex,
    scriptPath,
    simplifyText,
    splitForValues,
)



from __future__ import annotations

import json
import os
import re
from collections import defaultdict

import requests
import tenacity
from alive_progress import alive_it

from .. import config, t
from .. import messages as m

if t.TYPE_CHECKING:
    AnchorsT: t.TypeAlias = defaultdict[str, list["AnchorT"]]
    HeadingGroupT: t.TypeAlias = dict["SpecStatusT", "HeadingT"]
    HeadingKeyT: t.TypeAlias = str
    SpecStatusT: t.TypeAlias = str
    HeadingsT: t.TypeAlias = dict[HeadingKeyT, list[HeadingKeyT] | HeadingGroupT]
    AllHeadingsT: t.TypeAlias = dict[str, HeadingsT]
    SpecsT: t.TypeAlias = dict[str | None, "SpecT"]
    MethodsT: t.TypeAlias = defaultdict[str, dict[str, t.Any]]
    ForsT: t.TypeAlias = defaultdict[str, list[str]]

    # https://github.com/w3c/reffy/blob/main/schemas/browserlib/extract-headings.json
    class WebrefHeadingT(t.TypedDict, total=False):
        id: t.Required[str]
        href: t.Required[str]
        title: t.Required[str]
        level: t.Required[int]
        number: str
        alternateIds: list[str]

    class HeadingT(t.TypedDict):
        url: str
        number: str
        text: str
        spec: str

    # https://github.com/w3c/reffy/blob/main/schemas/browserlib/extract-dfns.json
    WebrefAnchorT = t.TypedDict(
        "WebrefAnchorT",
        {
            "href": t.Required[str],
            "linkingText": list[str],
            "type": t.Required[str],
            "for": t.Required[list[str]],
            "access": t.Required[str],
            "informative": t.Required[bool],
        },
        total=False,
    )

    AnchorT = t.TypedDict(
        "AnchorT",
        {
            "status": str,
            "type": str,
            "spec": str,
            "shortname": str,
            "level": int,
            "export": bool,
            "normative": bool,
            "url": str,
            "for": list[str],
        },
    )

    class WebrefSpecT(t.TypedDict, total=False):
        url: t.Required[str]
        shortname: t.Required[str]
        title: t.Required[str]
        shortTitle: t.Required[str]
        series: t.Required[dict[str, str]]
        nightly: t.Required[dict[str, str | list[str]]]
        release: dict[str, str | list[str]]
        seriesVersion: str
        seriesComposition: str
        links: str
        refs: str
        idl: str
        dfns: str
        headings: str
        ids: str

    class SpecT(t.TypedDict):
        vshortname: str
        shortname: str
        snapshot_url: str | None
        current_url: str | None
        title: str
        description: str | None
        abstract: str | None
        level: int | None


def update(path: str, dryRun: bool = False) -> set[str] | None:
    specs: SpecsT = {}
    anchors: AnchorsT = defaultdict(list)
    headings: AllHeadingsT = {}

    gatherWebrefData(specs, anchors, headings)
    cleanSpecHeadings(headings)
    methods = extractMethodData(anchors)
    fors = extractForsData(anchors)

    if not dryRun:
        writtenPaths = set()
        try:
            p = os.path.join(path, "specs.json")
            writtenPaths.add(p)
            with open(p, "w", encoding="utf-8") as f:
                f.write(json.dumps(specs, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            m.die(f"Couldn't save spec database to disk.\n{e}")
            return None
        try:
            for specName, specHeadings in headings.items():
                p = os.path.join(path, "headings", f"headings-{specName}.json")
                writtenPaths.add(p)
                with open(p, "w", encoding="utf-8") as f:
                    f.write(json.dumps(specHeadings, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            m.die(f"Couldn't save headings database to disk.\n{e}")
            return None
        try:
            writtenPaths.update(writeAnchorsFile(anchors, path))
        except Exception as e:
            m.die(f"Couldn't save anchor database to disk.\n{e}")
            return None
        try:
            p = os.path.join(path, "methods.json")
            writtenPaths.add(p)
            with open(p, "w", encoding="utf-8") as f:
                f.write(json.dumps(methods, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            m.die(f"Couldn't save methods database to disk.\n{e}")
            return None
        try:
            p = os.path.join(path, "fors.json")
            writtenPaths.add(p)
            with open(p, "w", encoding="utf-8") as f:
                f.write(json.dumps(fors, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            m.die(f"Couldn't save fors database to disk.\n{e}")
            return None

    m.say("Success!")
    return writtenPaths


def gatherWebrefData(specs: SpecsT, anchors: AnchorsT, headings: AllHeadingsT) -> None:
    m.say("Downloading anchor data from Webref...")
    currentWebrefData = [x for x in specsFromWebref("current") if "dfns" in x or "headings" in x]
    snapshotWebrefData = specsFromWebref("snapshot")

    progress = alive_it(currentWebrefData, dual_line=True)
    for rawWSpec in t.cast("t.Generator[WebrefSpecT, None, None]", progress):
        spec = canonSpecFromWebref(rawWSpec)
        currentUrl = spec["current_url"]
        assert currentUrl is not None
        specName = spec["vshortname"].lower()
        progress.text(specName)

        specs[specName] = spec
        specHeadings: HeadingsT = {}
        headings[spec["vshortname"]] = specHeadings

        if "dfns" in rawWSpec:
            currentAnchors = anchorsFromWebref("current", rawWSpec["dfns"])
            if currentAnchors:
                for anchor in currentAnchors:
                    addToAnchors(anchor, anchors, spec, "current")
        if "headings" in rawWSpec:
            currentHeadings = headingsFromWebref("current", rawWSpec["headings"])
            if currentHeadings:
                for heading in currentHeadings:
                    addToHeadings(heading, specHeadings, spec, "current")

        # Complete list of anchors/headings with those from the snapshot version of the spec
        if spec["snapshot_url"] is not None and snapshotWebrefData is not None:
            rawSnapshotSpec: WebrefSpecT | None = None
            for s in snapshotWebrefData:
                if s["shortname"].lower() == spec["vshortname"].lower():
                    rawSnapshotSpec = s
                    break
            else:
                m.warn(f"Despite claiming to have a snapshot url, no snapshot data found for '{spec['vshortname']}'.")
            if rawSnapshotSpec:
                if "dfns" in rawSnapshotSpec:
                    snapshotAnchors = anchorsFromWebref("snapshot", rawSnapshotSpec["dfns"])
                    if snapshotAnchors:
                        for anchor in snapshotAnchors:
                            addToAnchors(anchor, anchors, spec, "snapshot")
                if "headings" in rawSnapshotSpec:
                    snapshotHeadings = headingsFromWebref("snapshot", rawSnapshotSpec["headings"])
                    if snapshotHeadings:
                        for heading in snapshotHeadings:
                            addToHeadings(heading, specHeadings, spec, "snapshot")


def specsFromWebref(status: t.Literal["current" | "snapshot"]) -> list[WebrefSpecT]:
    url = ("ed" if status == "current" else "tr") + "/index.json"
    j = dataFromWebref(url)
    if j is None or j.get("results") is None:
        msg = f"No {status} specs data from WebRef. Got:\n{json.dumps(j, indent=1)}"
        raise Exception(msg)
    rawSpecs = t.cast("list[WebrefSpecT]", j["results"])
    filteredSpecs: list[WebrefSpecT] = []
    for spec in rawSpecs:
        if spec["seriesComposition"] == "fork":
            continue
        filteredSpecs.append(spec)
    return filteredSpecs


def anchorsFromWebref(status: t.Literal["current" | "snapshot"], urlSuffix: str) -> list[WebrefAnchorT]:
    url = ("ed" if status == "current" else "tr") + "/" + urlSuffix
    j = dataFromWebref(url)
    if j is None or j.get("dfns") is None:
        msg = f"No WebRef dfns data at {url}. Got:\n{json.dumps(j, indent=1)}"
        raise Exception(msg)
    return t.cast("list[WebrefAnchorT]", j["dfns"])


def headingsFromWebref(status: t.Literal["current" | "snapshot"], urlSuffix: str) -> list[WebrefHeadingT]:
    url = ("ed" if status == "current" else "tr") + "/" + urlSuffix
    j = dataFromWebref(url)
    if j is None or j.get("headings") is None:
        msg = f"No WebRef headings data at {url}. Got:\n{json.dumps(j, indent=1)}"
        raise Exception(msg)
    return t.cast("list[WebrefHeadingT]", j["headings"])


@tenacity.retry(reraise=True, stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(1, 2))
def dataFromWebref(url: str) -> t.JSONT:
    webrefAPIUrl = "https://raw.githubusercontent.com/w3c/webref/main/"
    try:
        response = requests.get(webrefAPIUrl + url, timeout=5)
    except Exception as e:
        msg = f"Couldn't download data from Webref.\n{e}"
        raise Exception(msg) from e
    try:
        data = response.json()
    except Exception as e:
        msg = f"Data retrieved from Webref wasn't valid JSON for some reason. Try downloading again?\n{e}"
        raise Exception(msg) from e
    return t.cast("t.JSONT", data)


def canonSpecFromWebref(rawSpec: WebrefSpecT) -> SpecT:
    """Generate a spec object from data gleaned from Webref"""
    assert rawSpec["shortname"] is not None
    assert rawSpec["series"] is not None
    assert rawSpec["nightly"] is not None
    assert rawSpec["title"] is not None
    assert rawSpec["shortTitle"] is not None
    spec: SpecT = {
        "vshortname": rawSpec["shortname"].lower(),
        "shortname": rawSpec["series"]["shortname"].lower(),
        "snapshot_url": t.cast("str", rawSpec.get("release", {}).get("url")),
        "current_url": t.cast("str", rawSpec["nightly"]["url"]),
        "title": rawSpec["shortTitle"],
        "description": rawSpec["title"],
        "abstract": None,
        "level": (
            int(rawSpec["seriesVersion"])
            if "seriesVersion" in rawSpec and re.match(r"^\d+$", rawSpec.get("seriesVersion", ""))
            else None
        ),
    }
    return spec


def addToAnchors(
    rawAnchor: WebrefAnchorT,
    anchors: AnchorsT,
    spec: SpecT,
    status: t.Literal["current"] | t.Literal["snapshot"],
) -> None:
    baseUrl = spec["snapshot_url"] if status == "snapshot" else spec["current_url"]
    assert baseUrl is not None
    assert rawAnchor["type"] is not None
    assert rawAnchor["linkingText"] is not None
    assert rawAnchor["access"] is not None
    assert rawAnchor["href"] is not None
    assert rawAnchor["informative"] is not None
    assert rawAnchor["for"] is not None

    anchor: AnchorT = {
        "status": status,
        "type": rawAnchor["type"],
        "export": rawAnchor["access"] == "public",
        "normative": not rawAnchor["informative"],
        "url": rawAnchor["href"],
        "for": rawAnchor["for"],
        "spec": spec["vshortname"],
        "shortname": spec["shortname"],
        "level": int(spec["level"] or "1"),
    }
    for text in rawAnchor["linkingText"]:
        text = re.sub(r"‘|’", "'", text)
        text = re.sub(r"“|”", '"', text)
        text = re.sub(r"\s+", " ", text)
        anchors[text].append(anchor)


def addToHeadings(
    rawAnchor: WebrefHeadingT,
    specHeadings: HeadingsT,
    spec: SpecT,
    status: t.Literal["current"] | t.Literal["snapshot"],
) -> None:
    heading: HeadingT = {
        "url": rawAnchor["href"],
        "number": rawAnchor.get("number", ""),
        "text": rawAnchor["title"],
        "spec": spec["title"],
    }
    specUrl = spec["snapshot_url"] if status == "snapshot" else spec["current_url"]
    assert specUrl is not None
    if heading["url"].startswith(specUrl):
        truncatedUrl = heading["url"][len(specUrl) :]
    else:
        m.warn(f"Invalid heading - URL doesn't start with the spec's url <{specUrl}>.\n{heading}")
        return
    if not truncatedUrl:
        m.warn(f"Invalid url - heading doesn't actually have an ID:\n{heading}")
        return
    if truncatedUrl[0] == "#":
        # Either single-page spec, or link on the top page of a multi-page spec
        fragment = truncatedUrl
        shorthand = "/" + fragment
    else:
        # Multi-page spec, need to guard against colliding IDs
        if "#" in truncatedUrl:
            # url to a heading in the page, like "foo.html#bar"
            match = re.match(r"([\w-]+).*?(#.*)", truncatedUrl)
            if not match:
                m.die(
                    f"Unexpected URI pattern '{truncatedUrl}' for spec '{spec['vshortname']}'. Please report this to the Bikeshed maintainer.",
                )
                return
            page, fragment = match.groups()
            page = "/" + page
        else:
            # url to a page itself, like "foo.html"
            page, _, _ = truncatedUrl.partition(".")
            page = "/" + page
            fragment = "#"
        shorthand = page + fragment

    def addShorthandAndFragmentToSpecHeadings(shorthand: str, fragment: str) -> None:
        if shorthand not in specHeadings:
            specHeadings[shorthand] = {}
        headingGroup = t.cast("HeadingGroupT", specHeadings[shorthand])
        headingGroup[status] = heading
        if fragment not in specHeadings:
            specHeadings[fragment] = []
        keyList = t.cast("list[HeadingKeyT]", specHeadings[fragment])
        if shorthand not in specHeadings[fragment]:
            keyList.append(shorthand)

    addShorthandAndFragmentToSpecHeadings(shorthand, fragment)

    # A heading may have multiple IDs: add the others as aliases of the primary one.
    if "alternateIds" in rawAnchor:
        shorthandBase, _, _ = shorthand.partition("#")
        for alternateId in rawAnchor["alternateIds"]:
            fragment = "#" + alternateId
            addShorthandAndFragmentToSpecHeadings(shorthandBase + fragment, fragment)


def cleanSpecHeadings(headings: AllHeadingsT) -> None:
    """Headings data was purposely verbose, assuming collisions even when there wasn't one.
    Want to keep the collision data for multi-page, so I can tell when you request a non-existent page,
    but need to collapse away the collision stuff for single-page."""
    for specHeadings in headings.values():
        for k, v in list(specHeadings.items()):
            if k[0] != "#":
                # a HeadingGroupT, not a list of HeadingKeyTs
                continue
            assert isinstance(v, list)
            if len(v) == 1 and v[0][0:2] == "/#":
                # No collision, and this is either a single-page spec or a non-colliding front-page link
                # Go ahead and collapse them.
                specHeadings[k] = specHeadings[v[0]]
                del specHeadings[v[0]]


def extractMethodData(anchors: AnchorsT) -> MethodsT:
    """Compile a db of {argless methods => {argfull method => {args, fors, url, shortname}}"""

    methods: defaultdict[str, dict] = defaultdict(dict)
    for key, anchors_ in anchors.items():
        # Extract the name and arguments
        match = re.match(r"([^(]+)\((.*)\)", key)
        if not match:
            continue
        methodName, argstring = match.groups()
        arglessMethod = methodName + "()"
        args = [x.strip() for x in argstring.split(",")] if argstring else []
        for anchor in anchors_:
            if anchor["type"] not in config.idlMethodTypes:
                continue
            if key not in methods[arglessMethod]:
                methods[arglessMethod][key] = {
                    "args": args,
                    "for": set(),
                    "shortname": anchor["shortname"],
                }
            methods[arglessMethod][key]["for"].update(anchor["for"])
    # Translate the "for" set back to a list for JSONing
    for signatures in methods.values():
        for signature in signatures.values():
            signature["for"] = sorted(signature["for"])
    return methods


def extractForsData(anchors: AnchorsT) -> ForsT:
    """Compile a db of {for value => dict terms that use that for value}"""

    fors: ForsT = defaultdict(list)
    for key, anchors_ in anchors.items():
        for anchor in anchors_:
            for for_ in anchor["for"]:
                if for_ == "":
                    continue
                fors[for_].append(key)
            if not anchor["for"]:
                fors["/"].append(key)
    for key, val in list(fors.items()):
        fors[key] = sorted(set(val))
    return fors


def writeAnchorsFile(anchors: AnchorsT, path: str) -> set[str]:
    """
    Keys may be duplicated.

    key
    type
    spec
    shortname
    level
    status
    url
    export (boolish string)
    normative (boolish string)
    for* (one per line, unknown #)
    - (by itself, ends the segment)
    """
    writtenPaths = set()
    groupedEntries: defaultdict[str, dict[str, list[AnchorT]]] = defaultdict(dict)
    for key, entries in anchors.items():
        group = config.groupFromKey(key)
        groupedEntries[group][key] = entries
    for group, group_anchors in groupedEntries.items():
        p = os.path.join(path, "anchors", f"anchors-{group}.data")
        writtenPaths.add(p)
        with open(p, "w", encoding="utf-8") as fh:
            for key, entries in sorted(group_anchors.items(), key=lambda x: x[0]):
                for e in sorted(entries, key=lambda x: x["url"]):
                    refKey, displayKey = config.adjustKey(key, e.get("type", ""))
                    fh.write(refKey + "\n")
                    fh.write(displayKey + "\n")
                    for field in [
                        "type",
                        "spec",
                        "shortname",
                        "level",
                        "status",
                        "url",
                    ]:
                        fh.write(str(e.get(field, "")) + "\n")
                    for field in ["export", "normative"]:
                        if e.get(field, False):
                            fh.write("1\n")
                        else:
                            fh.write("\n")
                    for forValue in e.get("for", []):
                        if forValue:  # skip empty strings
                            fh.write(forValue + "\n")
                    fh.write("-" + "\n")
    return writtenPaths



from __future__ import annotations

from .. import h, t
from .. import messages as m


def exampleIDs(doc: t.SpecT) -> None:
    """
    Checks that every example in the document has an ID.
    """
    if not doc.md.complainAbout["missing-example-ids"]:
        return
    for el in h.findAll(".example:not([id])", doc):
        m.lint(f"Example needs ID:\n{h.outerHTML(el)[0:100]}", el=el)



import enum


class UpdateMode(enum.Flag):
    """
    Enum dictating how the update process should run.
    """

    NONE = 0

    # Try to update via manifest
    MANIFEST = enum.auto()

    # Try to update manually
    MANUAL = enum.auto()

    # Update via the specified methods even if you'd normally skip
    # (such as if they would pass a freshness check).
    FORCE = enum.auto()

    BOTH = MANIFEST | MANUAL



from __future__ import annotations

from .. import h, t
from .. import messages as m

# Heading data for a spec is a dict.
# Each key is either an id, like "#id",
# or a page name + id, like "/page#id".
# The value is either a list of "/page#id" keys
# (when an #id appears in several pages in the spec),
# or a {"current": ..., "snapshot":...} dict,
# which can have either or both of these keys.
# Each of these latter dicts' values
# are a dict[str, str] of heading data like
# {
#   "number": "7.1",
#   "spec": "CSS Overflow 4",
#   "text": "Fragment styling",
#   "url": "https://drafts.csswg.org/css-overflow-4/#fragment-styling"
# }


class SpecHeadings:
    def __init__(self, spec: str, data: dict[str, list[str] | dict[str, dict[str, str]]]) -> None:
        self.spec = spec
        self.data = data

    def get(self, id: str, status: str, el: t.ElementT | None = None) -> dict[str, str] | None:
        currentHeading: dict[str, str] | None
        snapshotHeading: dict[str, str] | None
        if id in self.data:
            heading = self.data[id]
            if isinstance(heading, dict):
                currentHeading = heading.get("current")
                snapshotHeading = heading.get("snapshot")
            if isinstance(heading, list):
                clashingIds = heading
                # Multipage spec, list is "/page#id" keys
                # that (potentially) collide for that heading ID
                if len(clashingIds) == 1:
                    # only one heading of this name, no worries
                    x = self.data[clashingIds[0]]
                    assert isinstance(x, dict)
                    currentHeading = x.get("current")
                    snapshotHeading = x.get("snapshot")
                else:
                    # multiple headings of this id, user needs to disambiguate
                    m.die(
                        f"Multiple headings with id '{id}' for spec '{self.spec}'. Please specify:\n"
                        + "\n".join(f"  [[{self.spec + x}]]" for x in clashingIds),
                        el=el,
                    )
                    return None
        else:
            m.die(
                f"Couldn't find section '{id}' in spec '{self.spec}':\n{h.outerHTML(el)}",
                el=el,
            )
            return None
        if status == "current":
            return currentHeading or snapshotHeading
        else:
            return snapshotHeading or currentHeading
